from mpi4py import MPI
import sys

__mpimanager = None

class MPIManager(object):
    """Defines the behavior of the slaves/worker processes

    This class construct the MPI communicators structure needed
    if the rank of the process is in scheduler_node_ranks, the process is a scheduler
    then there is process_per_model process per communicator
    """

    def __init__(self, scheduler_node_ranks=[0], process_per_model=1):
        """
        Parameters
        ----------
        scheduler_node_ranks: Python list
            list of ranks computation should not happen on.
            Should include the scheduler so it doesn't get 
            overwhelmed with work.

        process_per_model: Integer
            the number of process to allow to each model
       """

        self._world_communicator = MPI.COMM_WORLD
        self._size = self._world_communicator.Get_size()
        self._rank = self._world_communicator.Get_rank()

        #Construct the appropriate communicators for resource allocation to models
        #There is one communicator for scheduler nodes
        #And one communicator per model
        self._scheduler_node_ranks = scheduler_node_ranks
        self._process_per_model = process_per_model
        self._model_color = int(((self._rank - sum(i < self._rank for i in scheduler_node_ranks)) / process_per_model) + 1)
        if(self._rank in scheduler_node_ranks):
            self._model_color = 0
        self._model_communicator = MPI.COMM_WORLD.Split(self._model_color, self._rank)
        self._model_size = self._model_communicator.Get_size()
        self._model_rank = self._model_communicator.Get_rank()

        # create a communicator to broadcast instructions to slaves
        self._scheduler_color = 1
        if(self._model_color == 0 or self._model_rank == 0):
            self._scheduler_color = 0
        self._scheduler_communicator = MPI.COMM_WORLD.Split(self._scheduler_color, self._rank)
        self._scheduler_size = self._scheduler_communicator.Get_size()
        self._scheduler_rank = self._scheduler_communicator.Get_rank()

        self._leader = False
        self._scheduler = False
        self._team = False
        self._worker = False

        if self._rank == 0:
            self._scheduler = True
        elif self._model_rank == 0:
            self._team = True
            self._leader = True
        else:
            self._team = True
            self._worker = True


    def is_scheduler(self):
        ''' Tells if the process is a scheduler '''
        return self._scheduler

    def is_team(self):
        ''' Tells if the process is a team '''
        return self._team

    def is_leader(self):
        ''' Tells if the process is a leader '''
        return self._leader

    def is_worker(self):
        ''' Tells if the process is a worker '''
        return self._worker

    def get_scheduler_node_ranks(self):
        ''' Returns the list of scheduler node wanks '''
        return self._scheduler_node_ranks

    def get_world_rank(self):
        ''' Returns the current rank '''
        return self._rank

    def get_world_size(self):
        ''' Returns the size of the world communicator '''
        return self._size

    def get_world_communicator(self):
        ''' Returns the world communicator '''
        return self._world_communicator

    def get_model_rank(self):
        ''' Returns the rank in the world communicator '''
        return self._model_rank

    def get_model_size(self):
        ''' Returns the size of the model communicator '''
        return self._model_size

    def get_model_communicator(self):
        ''' Returns the model communicator '''
        return self._model_communicator

    def get_scheduler_rank(self):
        ''' Returns the rank in the scheduler communicator '''
        return self._scheduler_rank

    def get_scheduler_size(self):
        ''' Returns the size of the scheduler communicator '''
        return self._scheduler_size

    def get_scheduler_communicator(self):
        ''' Returns the scheduler communicator '''
        return self._scheduler_communicator

def get_mpi_manager():
    ''' Return the instance of mpimanager
    Creates one with default parameters is not already existing '''
    global mpimanager
    if mpimanager == None :
        create_mpi_manager([0], 1)
    return mpimanager

def create_mpi_manager(scheduler_node_ranks, process_per_model):
    ''' Creates the instance of mpimanager with given parameters '''
    global mpimanager
    mpimanager = MPIManager(scheduler_node_ranks, process_per_model)


import unittest
import numpy as np

from abcpy.output import Journal

class JournalTests(unittest.TestCase):
    # def test_add_parameters(self):
    #     params1 = np.zeros((2,4))
    #     params2 = np.ones((2,4))
    #
    #     # test whether production mode only stores the last set of parameters
    #     journal_prod = Journal(0)
    #     journal_prod.add_parameters(params1)
    #     journal_prod.add_parameters(params2)
    #     self.assertEqual(len(journal_prod.parameters), 1)
    #     np.testing.assert_equal(journal_prod.parameters[0], params2)
    #
    #     # test whether reconstruction mode stores all parameter sets
    #     journal_recon = Journal(1)
    #     journal_recon.add_parameters(params1)
    #     journal_recon.add_parameters(params2)
    #     self.assertEqual(len(journal_recon.parameters), 2)
    #     np.testing.assert_equal(journal_recon.parameters[0], params1)
    #     np.testing.assert_equal(journal_recon.parameters[1], params2)



    def test_add_weights(self):
        weights1 = np.zeros((2,4))
        weights2 = np.ones((2,4))

        # test whether production mode only stores the last set of parameters
        journal_prod = Journal(0)
        journal_prod.add_weights(weights1)
        journal_prod.add_weights(weights2)
        self.assertEqual(len(journal_prod.weights), 1)
        np.testing.assert_equal(journal_prod.weights[0], weights2)

        # test whether reconstruction mode stores all parameter sets
        journal_recon = Journal(1)
        journal_recon.add_weights(weights1)
        journal_recon.add_weights(weights2)
        self.assertEqual(len(journal_recon.weights), 2)
        np.testing.assert_equal(journal_recon.weights[0], weights1)
        np.testing.assert_equal(journal_recon.weights[1], weights2)



    def test_add_opt_values(self):
        opt_values1 = np.zeros((2,4))
        opt_values2 = np.ones((2,4))

        # test whether production mode only stores the last set of parameters
        journal_prod = Journal(0)
        journal_prod.add_opt_values(opt_values1)
        journal_prod.add_opt_values(opt_values2)
        self.assertEqual(len(journal_prod.opt_values), 1)
        np.testing.assert_equal(journal_prod.opt_values[0], opt_values2)

        # test whether reconstruction mode stores all parameter sets
        journal_recon = Journal(1)
        journal_recon.add_opt_values(opt_values1)
        journal_recon.add_opt_values(opt_values2)
        self.assertEqual(len(journal_recon.opt_values), 2)
        np.testing.assert_equal(journal_recon.opt_values[0], opt_values1)
        np.testing.assert_equal(journal_recon.opt_values[1], opt_values2)



    def test_load_and_save(self):
        params1 = np.zeros((2,4))
        weights1 = np.zeros((2,4))

        journal = Journal(0)
        #journal.add_parameters(params1)
        journal.add_weights(weights1)
        journal.save('journal_tests_testfile.pkl')

        new_journal = Journal.fromFile('journal_tests_testfile.pkl')
        #np.testing.assert_equal(journal.parameters, new_journal.parameters)
        np.testing.assert_equal(journal.weights, new_journal.weights)
        




if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod
from numbers import Number
import numpy as np


class InputConnector():
    def __init__(self, dimension):
        """
        Creates input parameters of given dimensionality. Each dimension needs to be specified using the set method.

        Parameters
        ----------
        dimension: int
            Dimensionality of the input parameters.
        """

        self._all_indices_specified = False
        self._dimension = dimension
        self._models = [None]*dimension
        self._model_indices = [None]*dimension


    def from_number(number):
        """
        Convenient initializer that converts a number to a hyperparameter input parameter.

        Parameters
        ----------
        number

        Returns
        -------
        InputConnector
        """

        if isinstance(number, Number):
            input_parameters = InputConnector(1)
            input_parameters.set(0, Hyperparameter(number), 0)
            return input_parameters
        else:
            raise TypeError('Unsupported type.')


    def from_model(model):
        """
        Convenient initializer that converts the full output of a model to input parameters.

        Parameters
        ----------
        ProbabilisticModel

        Returns
        -------
        InputConnector
        """

        if isinstance(model, ProbabilisticModel):
            input_parameters = InputConnector(model.get_output_dimension())
            for i in range(model.get_output_dimension()):
                input_parameters.set(i, model, i)
            return input_parameters
        else:
            raise TypeError('Unsupported type.')


    def from_list(parameters):
        """
        Creates an InputParameters object from a list of ProbabilisticModels.

        In this case, number of input parameters equals the sum of output dimensions of all models in the parameter
        list. Further, the output and models are connected to the input parameters in the order they appear in the
        parameter list.

        For convenience,
        - the parameter list can contain nested lists
        - the method also accepts numbers instead of models, which are automatically converted to hyper parameters.

        Parameters
        ----------
        parameters: list
            A list of ProbabilisticModels

        Returns
        -------
        InputConnector
        """

        if isinstance(parameters, list):
            unnested_parameters = []
            parameters_count = 0
            for item in parameters:
                input_parameters_from_item = item
                if isinstance(item, list):
                    input_parameters_from_item = InputConnector.from_list(item)
                elif isinstance(item, (Hyperparameter, ProbabilisticModel)):
                    input_parameters_from_item = InputConnector.from_model(item)
                elif isinstance(item, Number):
                    input_parameters_from_item = InputConnector.from_number(item)
                elif not isinstance(item, InputConnector):
                    raise TypeError('Unsupported type.')

                unnested_parameters.append(input_parameters_from_item)
                parameters_count += input_parameters_from_item.get_parameter_count()

            # here, unnested_parameters is a list of InputConnector and parameters_count hold the total number of
            # parameters in this list
            input_parameters = InputConnector(parameters_count)
            index = 0
            for param in unnested_parameters:
                for pi in range(0, param.get_parameter_count()):
                    input_parameters.set(index, param._models[pi], param._model_indices[pi])
                    index += 1
            return input_parameters
        else:
            raise TypeError('Input is not a list')



    def __getitem__(self, index):
        """
        For the input models, return those fixed value(s) that are specified by index.

        In case a value of an input model is not (yet) fixed, None is returned.

        In other words, if the input models are interpreted as random variables, this method returns a realization for
        all random variables used as input parameter, which is specified by 'index'.

        Parameters
        ----------
        index: int, slice

        Returns
        -------
        int, float, list, None
        """

        # index is a single number
        if isinstance (index, Number):
            model = self._models[index]
            model_index = self._model_indices[index]
            if model.get_stored_output_values() is None:
                return None
            else:
                output_values = model.get_stored_output_values()
                return output_values[model_index]

        # index is a slice
        elif isinstance(index, slice):
            result = []
            for i in range(index.start, index.stop):
                val = self[i]
                if val != None:
                    result.append(val)
                else:
                    return None
            return result


    def get_values(self):
        """
        Returns the fixed values of all input models.

        Returns
        -------
        np.array
        """

        result = [0]*self._dimension
        for i in range(0, self._dimension):
            result[i] = self.__getitem__(i)
        return result


    def get_models(self):
        """
        Returns a list of all models.

        Returns
        -------
        list
        """

        return self._models


    def get_model(self, index):
        """
        Returns the model at index.

        Returns
        -------
        ProbabilisticModel
        """

        return self._models[index]


    def get_parameter_count(self):
        """
        Returns the number of parameters.

        Returns
        -------
        int
        """

        return self._dimension


    def set(self, index, model, model_index):
        """
        Sets for an input parameter index the input model and the model index to use.

        For convenience, model can also be a number, which is automatically casted to a hyper parameter.

        Parameters
        ----------
        index: int
            Index of the input parameter to be set.
        model: ProbabilisticModel, Number
            The model to be set for the input parameter.
        model_index: int
            Index of model's output to be used as input parameter.
        """

        if isinstance(model, Number):
            model = Hyperparameter(model)

        self._models[index] = model
        self._model_indices[index] = model_index
        if (self._models != None):
            self._all_indices_specified = True


    def all_models_fixed_values(self):
        """
        Checks whether all input models have fixed an output value (pseudo data).

        In order get a fixed output value (a realization of the random variable described by the model) a model has to
        run a forward simulation, which is not done automatically upon initialization.

        Returns
        -------
        boolean
        """

        for model in self._models:
            if model.get_stored_output_values() is None:
                return False
        return True


class ProbabilisticModel(metaclass = ABCMeta):
    """
    This abstract class represents all probabilistic models.
    """

    def __init__(self, input_connector, name=''):
        """
        This initializer *must be* called from any derived class to properly connect it to its input models.

        It accepts as input an InputConnector object that fully specifies how to connect all parent models to the
        current model.

        Parameters
        ----------
        input_connector: list
            A list of input parameters.

        name: string
            A human readable name for the model. Can be the variable name for example.
        """


        # set name
        self.name = name

        # parameters is of type InputConnector
        if isinstance(input_connector, InputConnector):
            if input_connector.all_models_fixed_values() and self._check_input(input_connector.get_values()) == False:
                raise ValueError('Input parameters are not compatible with current model.')
            self._parameters = input_connector
        else:
            raise TypeError('Input parameters are of wrong type.')

        self._fixed_values = None

        # A flag indicating whether the model has been touched during a recursive operation
        self.visited = False
        self.calculated_pdf = None


    def __getitem__(self, item):
        """
        Overloads the access operator. If the access operator is called, a tupel of the ProbablisticModel that called
        the operator and the index at which it was called is returned. Commonly used at initialization of new
        probabilistic models to specify a mapping between model outputs and parameters.

        Parameters
        ----------
        item: integer
            The index in the output of the parent model which should be linked to the parameter being defined.
        """

        if isinstance(item, Number):
            if(item>=self.get_output_dimension()):
                raise IndexError('The specified index lies out of range for probabilistic model %s.'%(self.__class__.__name__))
            input_parameters = InputConnector(1)
            input_parameters.set(0, self, item)
            return input_parameters
        else:
            raise TypeError('Input of unsupported type.')


    def get_input_values(self):
        """
        Returns the input values from the parent models as a list.
        Commonly used when sampling from the distribution.

        Returns
        -------
        list
        """

        return self.get_input_connector().get_values()


    def get_input_models(self):
        """
        Returns a list of all input models.

        Returns
        -------
        list
        """

        input_connector = self.get_input_connector()
        return input_connector.get_models()


    def get_stored_output_values(self):
        """
        Returns the stored sampled value of the probabilistic model after setting the values explicitly.

        At initialization the function should return None.

        Returns
        -------
        numpy.array or None.
        """

        return self._fixed_values


    def get_input_connector(self):
        """
        Returns the input connector object that connecects the current model to its parents.

        In case of no dependencies, this function should return None.

        Returns
        -------
        InputConnector, None
        """

        return self._parameters


    def get_input_dimension(self):
        """
        Returns the input dimension of the current model.

        Returns
        -------
        int

        """

        return self._parameters._dimension


    def set_output_values(self, values):
        """
        Sets the output values of the model. This method is commonly used to set new values after perturbing the old
        ones.


        Parameters
        ----------
        values: numpy array or dimension equal to output dimension.

        Returns
        -------
        boolean
            Returns True if it was possible to set the values, false otherwise.
        """
        if not isinstance(values, np.ndarray):
            raise TypeError('Elements of input list are not of type numpy array.')
        if values.shape[0] != self.get_output_dimension():
            raise IndexError('Size of input list not equal to number output dimensions.')

        if self._check_output(values):
            self._fixed_values = values
            return True
        return False


    def __add__(self, other):
        """Overload the + operator for probabilistic models.

        Parameters
        ----------
        other: probabilistic model or Hyperparameter
            The model to be added to self.

        Returns
        -------
        SummationModel
            A probabilistic model describing a model coming from summation.
        """
        return SummationModel([self,other])


    def __radd__(self, other):
        """Overload the + operator from the righthand side to support addition of Hyperparameters from the left.

        Parameters
        ----------
        Other: Hyperparameter
            The hyperparameter to be added to self.

        Returns
        -------
        SummationModel
            A probabilistic model describing a model coming from summation.
        """
        return SummationModel([other, self])


    def __sub__(self, other):
        """Overload the - operator for probabilistic models.

        Parameters
        ----------
        other: probabilistic model or Hyperparameter
            The model to be subtracted from self.

        Returns
        -------
        SubtractionModel
            A probabilistic model describing a model coming from subtraction.
        """
        return SubtractionModel([self, other])


    def __rsub__(self, other):
        """Overload the - operator from the righthand side to support subtraction of Hyperparameters from the left.

        Parameters
        ----------
        Other: Hyperparameter
            The hyperparameter to be subtracted from self.

        Returns
        -------
        SubtractionModel
            A probabilistic model describing a model coming from subtraction.
        """
        return SubtractionModel([other,self])


    def __mul__(self, other):
        """Overload the * operator for probabilistic models.

        Parameters
        ----------
        other: probabilistic model or Hyperparameter
            The model to be multiplied with self.

        Returns
        -------
        MultiplicationModel
            A probabilistic model describing a model coming from multiplication.
        """
        return MultiplicationModel([self,other])


    def __rmul__(self, other):
        """Overload the * operator from the righthand side to support subtraction of Hyperparameters from the left.

                Parameters
                ----------
                Other: Hyperparameter
                    The hyperparameter to be subtracted from self.

                Returns
                -------
                MultiplicationModel
                    A probabilistic model describing a model coming from multiplication.
                """
        return MultiplicationModel([other,self])


    def __truediv__(self, other):
        """Overload the / operator for probabilistic models.

        Parameters
        ----------
        other: probabilistic model or Hyperparameter
            The model to be divide self.

        Returns
        -------
        DivisionModel
            A probabilistic model describing a model coming from division.
        """
        return DivisionModel([self, other])


    def __rtruediv__(self, other):
        """Overload the / operator from the righthand side to support subtraction of Hyperparameters from the left.

        Parameters
        ----------
        Other: Hyperparameter
            The hyperparameter to be subtracted from self.

        Returns
        -------
        DivisionModel
            A probabilistic model describing a model coming from division.
        """
        return DivisionModel([other, self])


    def __pow__(self, power, modulo=None):
        return ExponentialModel([self, power])


    def __rpow__(self, other):
        return RExponentialModel([other, self])


    def _forward_simulate_and_store_output(self, rng=np.random.RandomState()):
        """
        Samples from the model associated and assigns the result to fixed_values, if applicable. Commonly used when
        sampling from the prior.

        Parameters
        ----------
        rng: Random number generator
            Defines the random number generator to be used by the sampling function.

        Returns
        -------
        boolean
            Check whether it was possible to set the parameters to sampled values.
        """

        parameters_are_valid = self._check_input(self.get_input_values())
        if(parameters_are_valid):
            sample_result = self.forward_simulate(self.get_input_values(), 1, rng=rng)
            if sample_result != None:
                self.set_output_values(sample_result[0])
                return True
        return False


    def pdf(self, input_values, x):
        """
        Calculates the probability density function at point x.

        Commonly used to determine whether perturbed parameters are still valid according to the pdf.

        Parameters
        ----------
        input_values: list
            List of input parameters, in the same order as specified in the InputConnector passed to the init function
        x: list
            The point at which the pdf should be evaluated.

        Returns
        -------
        float:
            The pdf evaluated at point x.
        """
        # If the probabilistic model is discrete, there is no probability density function, but a probability mass function. This check ensures that calling the pdf of such a model still works.
        if(isinstance(self, Discrete)):
            return self.pmf(input_values, x)
        else:
            raise NotImplementedError


    def calculate_and_store_pdf_if_needed(self, x):
        """
        Calculates the probability density function at point x and stores the result internally for later use.

        This function is intended to be used within the inference computation.

        Parameters
        ----------
        x: list
            The point at which the pdf should be evaluated.
        """

        if self._calculated_pdf == None:
            self._calculated_pdf = self.pdf(self.get_input_values(), x)


    def flush_stored_pdf(self):
        """
        This function flushes the internally stored value of a previously computed pdf.
        """

        self._calculated_pdf = None


    def get_stored_pdf(self):
        """
        Retrieves the value of a previously calculated pdf.

        Returns
        -------
        number
        """

        return self._calculated_pdf


    @abstractmethod
    def _check_input(self, input_values):
        """
        Check whether the input parameters are compatible with the underlying model.

        The following behavior is expected:

        1. If the input is of wrong type or has the wrong format, this method should *raise an exception*. For example,
        if the number of parameters does not match what the model expects.

        2. If the values of the input models are not compatible, this method should return False. For example, if an
        input value is not from the expected domain.

        Background information: Many inference schemes modify the input slightly by applying a small perturbation during
        sampling. This method is called to check whether the perturbation yields a reasonable input to the current
        model. In case this function returns False, the inference schemes re-perturb the input and try again. If the
        check is not done properly, the inference computation might **crash** or **not terminate**.

        Parameters
        ----------
        input_values: list
            A list of numbers that are the concatenation of all parent model outputs in the order specified by the
            InputConnector object that was passed during initialization.

        Returns
        -------
        boolean
            True if the fixed value of the parameters can be used as input for the current model. False otherwise.
        """

        raise NotImplementedError


    @abstractmethod
    def _check_output(self, values):
        """
        Checks whether values contains a reasonable output of the current model.

        Parameters
        ----------
        values: numpy array
            Array of shape (get_output_dimension(),) that contains the model output.

        Returns
        -------
        boolean
            Return false if values cannot possibly be generated from the model and true otherwise.
        """

        raise NotImplementedError


    @abstractmethod
    def forward_simulate(self, input_values, k, rng, mpi_comm=None):
        """
        Provides the output (pseudo data) from a forward simulation of the current model.

        In case the model is intended to be used as input for another model, a forward simulation **must** return a list
        of k numpy arrays with shape (get_output_dimension(),).

        In case the model is directly used for inference, and not as input for another model, a forward simulation
        also must return a list, but the elements can be arbitrarily defined. In this case it is only important that the
        used statistics and distance functions can read the input.

        Parameters
        ----------
        input_values: list
            A list of numbers that are the concatenation of all parent model outputs in the order specified by the
            InputConnector object that was passed during initialization.
        k: integer
            The number of forward simulations that should be run
        rng: Random number generator
            Defines the random number generator to be used. The default value uses a random seed to initialize the
            generator.
        mpi_comm: MPI communicator object
        	    Defines the MPI communicator object for MPI parallelization. The default value is None,
        	    meaning the forward simulation is not MPI-parallelized.

        Returns
        -------
        list
            A list of *k* elements, where each element is of type numpy arary and represents the result of a single
            forward simulation.
        """

        raise NotImplementedError


    @abstractmethod
    def get_output_dimension(self):
        """
        Provides the output dimension of the current model.

        This function is in particular important if the current model is used as an input for other models. In such a
        case it is assumed that the output is always a vector of int or float. The length of the vector is the dimension
        that should be returned here.

        Returns
        -------
        int:
            The dimension of the output vector of a single forward simulation.
        """

        raise NotImplementedError


class Continuous(metaclass = ABCMeta):
    """
    This abstract class represents all continuous probabilistic models.
    """

    @abstractmethod
    def pdf(self, input_values, x):
        """
        Calculates the probability density function of the model.

        Parameters
        ----------
        input_values: list
            A list of numbers that are the concatenation of all parent model outputs in the order specified by the
            InputConnector object that was passed during initialization.
        x: float
            The location at which the probability density function should be evaluated.
        """

        raise NotImplementedError


class Discrete(metaclass = ABCMeta):
    """
    This abstract class represents all discrete probabilistic models.
    """

    @abstractmethod
    def pmf(self, input_values, x):
        """
        Calculates the probability mass function of the model.

        Parameters
        ----------
        input_values: list
            A list of numbers that are the concatenation of all parent model outputs in the order specified by the
            InputConnector object that was passed during initialization.
        x: float
            The location at which the probability mass function should be evaluated.
        """

        raise NotImplementedError


class Hyperparameter(ProbabilisticModel):
    """
    This class represents all hyperparameters (i.e. fixed parameters).

    """
    def __init__(self, value, name='Hyperparameter'):
        """

        Parameters
        ----------
        value: list
            The values to which the hyperparameter should be set
        """

        # A hyperparameter is defined by the fact that it does not have any parents
        self.name = name
        self._fixed_values = np.array([value])
        self.visited = False


    def _forward_simulate_and_store_output(self, rng=np.random.RandomState()):
        self.visited = True
        return True


    def _check_input(self, input_values):
        """
        Hyperparameters have no input, thus we only accept None.
        """

        if not isinstance(input_values, list):
            raise TypeError
        if len(input_values) == 0:
            return True
        return False


    def _check_output(self, values):
        return False


    def set_output_values(self, values, rng=np.random.RandomState()):
        if not isinstance(values, np.ndarray):
            raise TypeError('Input not a numpy.array.')
        if values.shape[0] != 1:
            raise IndexError('Dimensions not matching.')
        return False


    def get_input_dimension(self):
        return 0;

    def get_output_dimension(self):
        return 1;


    def get_input_connector(self):
        return None


    def get_input_models(self):
        return []


    def get_input_values(self):
        return []


    def forward_simulate(self, input_values, k, rng=np.random.RandomState(), mpi_comm=None):
        return [np.array(self._fixed_values) for _ in range(k)]


    def pdf(self, input_values, x):
        # Mathematically, the expression for the pdf of a hyperparameter should be: if(x==self.fixed_parameters) return
        # 1; else return 0; However, since the pdf is called recursively for the whole model structure, and pdfs
        # multiply, this would mean that all pdfs become 0. Setting the return value to 1 ensures proper calulation of
        # the overall pdf.
        return 1.


class ModelResultingFromOperation(ProbabilisticModel):
    """This class implements probabilistic models returned after performing an operation on two probabilistic models
        """

    def __init__(self, parameters, name=''):
        """
        Parameters
        ----------
        parameters: list
            List containing two probabilistic models that should be added together.
        """

        if len(parameters) != 2:
            raise TypeError('Input list does not contain two models.')

        # here, parameters contains exactly two elements
        model_output_dim = [0, 0]
        for i, model in enumerate(parameters):
            if isinstance(model, ProbabilisticModel):
                model_output_dim[i] = model.get_output_dimension()
            elif isinstance(model, Number):
                model_output_dim[i] = 1
            else:
                raise TypeError('Unsupported type.')

        # here, model_output_dim contains the dim of both input models
        if model_output_dim[0] != model_output_dim[1]:
            raise ValueError('The provided models are not of equal dimension.')

        self._dimension = 1
        input_parameters = InputConnector.from_list(parameters)
        super(ModelResultingFromOperation, self).__init__(input_parameters, name)


    def forward_simulate(self, input_values, k, rng=np.random.RandomState(), mpi_comm=None):
        raise NotImplementedError


    def _check_input(self, input_values):
        return True


    def _check_output(self, parameters):
        """Checks parameters while setting them. Provided due to inheritance."""
        return True


    def get_output_dimension(self):
        return self._dimension


    def pdf(self, input_values, x):
        """Calculates the probability density function at point x.

        Parameters
        ----------
        input_values: list
            List of input parameters, in the same order as specified in the InputConnector passed to the init function
        x: float or list
            The point at which the pdf should be evaluated.

        Returns
        -------
        float
            The probability density function evaluated at point x.
        """
        # Since the nodes provided as input have to be independent, the resulting pdf will be pdf(parent 1)*pfd(parent 2). During the recursive graph action, this is calculated automatically, so the pdf at this node is expected to be 1
        return 1.

    def sample_from_input_models(self, k, rng=np.random.RandomState()):
        """
        Return for each input model k samples.

        Parameters
        ----------
        k: int
            Specifies the number of samples to generate from each input model.

        Returns
        -------
        dict
            A dictionary of type ProbabilisticModel:[], where the list contains k samples of the corresponding model.
        """

        model_samples = {}

        # Store the visited state of all input models
        visited_state = [False] * self.get_input_dimension()
        for i in range(0, self.get_input_dimension()):
            visited_state[i] = self.get_input_connector().get_model(i).visited

        # Set visited flag of all input models to False
        for i in range(0, self.get_input_dimension()):
            self.get_input_connector().get_model(i).visited = False

        # forward simulate each input model to get fixed input for the current model
        for i in range(0, self.get_input_dimension()):
            model = self.get_input_connector().get_model(i)
            if not model.visited:
                model_has_valid_parameters = model._check_input(model.get_input_values())
                if model_has_valid_parameters:
                    model_samples[model] = model.forward_simulate(model.get_input_values(), k, rng=rng)
                    model.visited = True
                else:
                    raise ValueError('Model %s has invalid input parameters.' % parent.name)

        # Restore the visited state of all input models
        for i in range(0, self.get_input_dimension()):
            self.get_input_connector().get_model(i).visited = visited_state[i]

        return model_samples


class SummationModel(ModelResultingFromOperation):
    """This class represents all probabilistic models resulting from an addition of two probabilistic models"""

    def forward_simulate(self, input_values, k, rng=np.random.RandomState(), mpi_comm=None):
        """Adds the sampled values of both parent distributions.

        Parameters
        ----------
        input_values: list
            List of input values
        k: integer
            The number of samples that should be sampled
        rng: random number generator
            The random number generator to be used.
        mpi_comm: MPI communicator object
            Defines the MPI communicator object for MPI parallelization. The default value is None,
            meaning the forward simulation is not MPI-parallelized.

        Returns
        -------
        list:
            The first entry is True, it is always possible to sample, given two parent values. The second entry is the
            sum of the parents values.
        """
        return_value = []

        # we need to obtain new samples of the parents for each sample (if we just use get_input_values, we will
        # have k identical samples)
        model_samples = self.sample_from_input_models(k, rng)

        for i in range(k):
            parameter_values = [0 for i in range(self.get_input_dimension())]
            for j in range(0, self.get_input_dimension()):
                model = self.get_input_connector().get_model(j)
                parameter_values[j] = model_samples[model][i]

            # add the corresponding parameter_values
            sample_value = []
            for j in range(self.get_output_dimension()):
                sample_value.append(parameter_values[j]+parameter_values[j+self.get_output_dimension()])
            if(len(sample_value)==1):
                sample_value=sample_value[0]
            return_value.append(sample_value)

        return return_value


class SubtractionModel(ModelResultingFromOperation):
    """This class represents all probabilistic models resulting from an subtraction of two probabilistic models"""

    def forward_simulate(self, input_values, k, rng=np.random.RandomState(), mpi_comm=None):
        """Adds the sampled values of both parent distributions.

        Parameters
        ----------
        input_values: list
            List of input values
        k: integer
            The number of samples that should be sampled
        rng: random number generator
            The random number generator to be used.
        mpi_comm: MPI communicator object
            Defines the MPI communicator object for MPI parallelization. The default value is None,
            meaning the forward simulation is not MPI-parallelized.

        Returns
        -------
        list:
            The first entry is True, it is always possible to sample, given two parent values. The second entry is the
            difference of the parents values.
        """
        return_value = []
        sample_value = []

        model_samples = self.sample_from_input_models(k, rng)

        for i in range(k):
            parameter_values = [0 for i in range(self.get_input_dimension())]
            for j in range(0, self.get_input_dimension()):
                model = self.get_input_connector().get_model(j)
                parameter_values[j] = model_samples[model][i]

            # subtract the corresponding parameter_values
            sample_value = []
            for j in range(self.get_output_dimension()):
                sample_value.append(parameter_values[j] - parameter_values[j + self.get_output_dimension()])
            if(len(sample_value)==1):
                sample_value=sample_value[0]
            return_value.append(sample_value)

        return return_value


class MultiplicationModel(ModelResultingFromOperation):
    """This class represents all probabilistic models resulting from a multiplication of two probabilistic models"""
    def forward_simulate(self, input_values, k, rng=np.random.RandomState(), mpi_comm=None):
        """Multiplies the sampled values of both parent distributions element wise.

        Parameters
        ----------
        input_values: list
            List of input values
        k: integer
            The number of samples that should be sampled
        rng: random number generator
            The random number generator to be used.
        mpi_comm: MPI communicator object
            Defines the MPI communicator object for MPI parallelization. The default value is None,
            meaning the forward simulation is not MPI-parallelized.

        Returns
        -------
        list:
            The first entry is True, it is always possible to sample, given two parent values. The second entry is the product of the parents values.
        """
        return_value = []

        model_samples = self.sample_from_input_models(k, rng)

        for i in range(k):
            parameter_values = [0 for i in range(self.get_input_dimension())]
            for j in range(0, self.get_input_dimension()):
                model = self.get_input_connector().get_model(j)
                parameter_values[j] = model_samples[model][i]

            # multiply the corresponding parameter_values
            sample_value = []

            for j in range(self.get_output_dimension()):
                sample_value.append(parameter_values[j] * parameter_values[j+self.get_output_dimension()])
            if (len(sample_value) == 1):
                sample_value = sample_value[0]
            return_value.append(sample_value)

        return return_value


class DivisionModel(ModelResultingFromOperation):
    """This class represents all probabilistic models resulting from a division of two probabilistic models"""

    def forward_simulate(self, input_valus, k, rng=np.random.RandomState(), mpi_comm=None):
        """Divides the sampled values of both parent distributions.

        Parameters
        ----------
        input_values: list
            List of input values
        k: integer
            The number of samples that should be sampled
        rng: random number generator
            The random number generator to be used.
        mpi_comm: MPI communicator object
            Defines the MPI communicator object for MPI parallelization. The default value is None,
            meaning the forward simulation is not MPI-parallelized.

        Returns
        -------
        list:
            The first entry is True, it is always possible to sample, given two parent values. The second entry is the fraction of the parents values.
        """
        return_value = []

        model_samples = self.sample_from_input_models(k, rng)

        for i in range(k):
            parameter_values = [0 for i in range(self.get_input_dimension())]
            for j in range(0, self.get_input_dimension()):
                model = self.get_input_connector().get_model(j)
                parameter_values[j] = model_samples[model][i]

            # divide the corresponding parameter_values
            sample_value = []

            for j in range(self.get_output_dimension()):
                sample_value.append(parameter_values[j]/parameter_values[j + self.get_output_dimension()])
            return_value += sample_value

        return return_value


class ExponentialModel(ModelResultingFromOperation):
    """This class represents all probabilistic models resulting from an exponentiation of two probabilistic models"""

    def __init__(self, parameters, name=''):
        """
        Specific initializer for exponential models that does additional checks.

        Parameters
        ----------
        parameters: list
            List of probabilistic models that should be added together.
        """

        exp = parameters[1]
        if isinstance(exp, ProbabilisticModel):
            if exp.get_output_dimension() != 1:
                raise ValueError('The exponent can only be 1 dimensional.')

        super(ExponentialModel, self).__init__(parameters, name)


    def _check_input(self, input_values):
        return True


    def forward_simulate(self, input_values, k, rng=np.random.RandomState(), mpi_comm=None):
        """Raises the sampled values of the base by the exponent.

        Parameters
        ----------
        input_values: list
            List of input values
        k: integer
            The number of samples that should be sampled
        rng: random number generator
            The random number generator to be used.
        mpi_comm: MPI communicator object
            Defines the MPI communicator object for MPI parallelization. The default value is None,
            meaning the forward simulation is not MPI-parallelized.

        Returns
        -------
        list:
            The first entry is True, it is always possible to sample, given two parent values. The second entry is the exponential of the parents values.
        """
        result = []

        model_samples = self.sample_from_input_models(k, rng)

        for i in range(k):
            parameter_values = [0 for i in range(self.get_input_dimension())]
            for j in range(0, self.get_input_dimension()):
                model = self.get_input_connector().get_model(j)
                parameter_values[j] = model_samples[model][i]

            power = parameter_values[-1]

            sample_value = []

            for j in range(self.get_output_dimension()):
                sample_value.append(parameter_values[j]**power)
            result.append(np.array(sample_value))

        return result


class RExponentialModel(ModelResultingFromOperation):
    """This class represents all probabilistic models resulting from an exponentiation of a Hyperparameter by another probabilistic model."""

    def __init__(self, parameters, name=''):
        """
        Specific initializer for exponential models that does additional checks.

        Parameters
        ----------
        parameters: list
            List of probabilistic models that should be added together.
        """

        exp = parameters[1]
        if isinstance(exp, ProbabilisticModel):
            if exp.get_output_dimension() != 1:
                raise ValueError('The exponent can only be 1 dimensional.')
        super(RExponentialModel, self).__init__(parameters, name)


    def _check_input(self, input_values):
        return True


    def forward_simulate(self, input_values, k, rng=np.random.RandomState(), mpi_comm=None):
        """Raises the base by the sampled value of the exponent.

        Parameters
        ----------
        input_values: list
            List of input values
        k: integer
            The number of samples that should be sampled
        rng: random number generator
            The random number generator to be used.
        mpi_comm: MPI communicator object
            Defines the MPI communicator object for MPI parallelization. The default value is None,
            meaning the forward simulation is not MPI-parallelized.

        Returns
        -------
        list:
            The first entry is True, it is always possible to sample, given two parent values. The second entry is the exponential of the parents values.
        """
        result = []

        model_samples = self.sample_from_input_models(k, rng)

        for i in range(k):
            parameter_values = [0 for i in range(self.get_input_dimension())]
            for j in range(0, self.get_input_dimension()):
                model = self.get_input_connector().get_model(j)
                parameter_values[j] = model_samples[model][i]

            power = parameter_values[0]

            sample_value = []

            for j in range(self.get_output_dimension()):
                sample_value.append(parameter_values[j] ** power)
            result.append(sample_value)

        return [np.array(result)]




import unittest
from abcpy.continuousmodels import Uniform
from abcpy.continuousmodels import Normal
from abcpy.continuousmodels import StudentT
from abcpy.statistics import Identity
from abcpy.backends import BackendDummy as Backend
from abcpy.modelselections import RandomForest
     
class RandomForestTests(unittest.TestCase):
    def setUp(self):
        # define observation for true parameters mean=170, std=15
        self.y_obs = [160.82499176]
        self.model_array = [None]*2
        #Model 1: Gaussian
        # define prior
        self.mu1 = Uniform([[150], [200]], name='mu1')
        self.sigma1 = Uniform([[5.0], [25.0]], name='sigma1')
        # define the model
        self.model_array[0] = Normal([self.mu1, self.sigma1])
        #Model 2: Student t
        # define prior
        self.mu2 = Uniform([[150], [200]], name='mu2')
        self.sigma2 = Uniform([[1], [30.0]], name='sigma2')
        # define the model
        self.model_array[1] = StudentT([self.mu2, self.sigma2])

        # define statistics
        self.statistics_calc = Identity(degree = 2, cross = False)
        # define backend
        self.backend = Backend()


    def test_select_model(self):
        modelselection = RandomForest(self.model_array, self.statistics_calc, self.backend, seed = 1)
        model = modelselection.select_model(self.y_obs,n_samples = 100, n_samples_per_param = 1)

        self.assertTrue(self.model_array[0] == model)
    
    def test_posterior_probability(self):
        modelselection = RandomForest(self.model_array, self.statistics_calc, self.backend, seed = 1)
        model_prob = modelselection.posterior_probability(self.y_obs)

        self.assertTrue(model_prob > 0.7)
 
if __name__ == '__main__':
    unittest.main()


from functools import wraps


def cached(func):
    cache = {}

    @wraps(func)
    def wrapped(x):
        if x not in cache:
            cache[x] = func(x)
        return cache[x]

    return wrapped



import unittest
import numpy as np

from abcpy.backends import BackendDummy
from abcpy.continuousmodels import Normal

from abcpy.distances import Euclidean

from abcpy.approx_lhd import SynLikelihood

from abcpy.continuousmodels import Uniform

from abcpy.statistics import Identity

from abcpy.inferences import RejectionABC, PMC, PMCABC, SABC, ABCsubsim, SMCABC, APMCABC, RSMCABC

class RejectionABCTest(unittest.TestCase):
    def test_sample(self):
        # setup backend
        dummy = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu,sigma])

        # define sufficient statistics for the model
        stat_calc = Identity(degree=2, cross=0)
        
        # define a distance function
        dist_calc = Euclidean(stat_calc)

        # create fake observed data
        y_obs = [np.array(9.8)]

        # use the rejection sampling scheme
        sampler = RejectionABC([self.model], [dist_calc], dummy, seed = 1)
        journal = sampler.sample([y_obs], 10, 1, 10)
        mu_sample = np.array(journal.get_parameters()['mu'])
        sigma_sample = np.array(journal.get_parameters()['sigma'])

        # test shape of samples
        mu_shape, sigma_shape = (len(mu_sample), mu_sample[0].shape[1]), \
                                                                    (len(sigma_sample),
                                                                     sigma_sample[0].shape[1])
        self.assertEqual(mu_shape, (10,1))
        self.assertEqual(sigma_shape, (10,1))

        # Compute posterior mean
        #self.assertAlmostEqual(np.average(np.asarray(samples[:,0])),1.22301,10e-2)
        self.assertLess(np.average(mu_sample) - 1.22301, 1e-2)
        self.assertLess(np.average(sigma_sample) - 6.992218,10e-2)

        self.assertFalse(journal.number_of_simulations==0)




class PMCTests(unittest.TestCase):
        
    def test_sample(self):
        # setup backend
        backend = BackendDummy()
        
        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu,sigma])

        # define sufficient statistics for the model
        stat_calc = Identity(degree = 2, cross = 0)

        # create fake observed data
        #y_obs = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        y_obs = [np.array(9.8)]
      
        # Define the likelihood function
        likfun = SynLikelihood(stat_calc)


        T, n_sample, n_samples_per_param = 1, 10, 100
        sampler = PMC([self.model], [likfun], backend, seed = 1)
        journal = sampler.sample([y_obs], T, n_sample, n_samples_per_param, covFactors =  np.array([.1,.1]), iniPoints = None)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))
        self.assertLess(abs(mu_post_mean - (-3.3711206204663764)), 1e-3)
        self.assertLess(abs(sigma_post_mean - 6.518520667688998), 1e-3)

        self.assertFalse(journal.number_of_simulations == 0)


        # use the PMC scheme for T = 2
        T, n_sample, n_samples_per_param = 2, 10, 100
        sampler = PMC([self.model], [likfun], backend, seed = 1)
        journal = sampler.sample([y_obs], T, n_sample, n_samples_per_param, covFactors = np.array([.1,.1]), iniPoints = None)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))
        self.assertLess(abs(mu_post_mean - (-2.970827684425406) ), 1e-3)
        self.assertLess(abs(sigma_post_mean - 6.82165619013458), 1e-3)

        self.assertFalse(journal.number_of_simulations == 0)


class PMCABCTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=0)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        #self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]

        
    def test_calculate_weight(self):
        n_samples = 2
        rc = PMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        theta = np.array([1.0,1.0])


        weight = rc._calculate_weight(theta)
        self.assertEqual(weight, 0.5)
        
        accepted_parameters = [[1.0, 1.0 + np.sqrt(2)],[0,0]]
        accepted_weights = np.array([[.5], [.5]])
        accepted_cov_mat = [np.array([[1.0,0],[0,1]])]
        rc.accepted_parameters_manager.update_broadcast(rc.backend, accepted_parameters, accepted_weights, accepted_cov_mat)
        kernel_parameters = []
        for kernel in rc.kernel.kernels:
            kernel_parameters.append(
                rc.accepted_parameters_manager.get_accepted_parameters_bds_values(kernel.models))

        rc.accepted_parameters_manager.update_kernel_values(rc.backend, kernel_parameters=kernel_parameters)
        weight = rc._calculate_weight(theta)
        expected_weight = 0.170794684453
        self.assertAlmostEqual(weight, expected_weight)
        

        
    def test_sample(self):
        # use the PMCABC scheme for T = 1
        T, n_sample, n_simulate, eps_arr, eps_percentile = 1, 10, 1, [10], 10
        sampler = PMCABC([self.model], [self.dist_calc], self.backend, seed = 1)
        journal = sampler.sample([self.observation], T, eps_arr, n_sample, n_simulate, eps_percentile)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())
          
        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))
        self.assertLess(mu_post_mean - 0.03713, 10e-2)
        self.assertLess(sigma_post_mean - 7.727, 10e-2)

        #self.assertEqual((mu_post_mean, sigma_post_mean), (,))
        
        # use the PMCABC scheme for T = 2
        T, n_sample, n_simulate, eps_arr, eps_percentile = 2, 10, 1, [10,5], 10
        sampler = PMCABC([self.model], [self.dist_calc], self.backend, seed = 1)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        journal = sampler.sample([self.observation], T, eps_arr, n_sample, n_simulate, eps_percentile)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))
        self.assertLess(mu_post_mean - 0.9356, 10e-2)
        self.assertLess(sigma_post_mean - 7.819, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)


class SABCTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=0)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        #self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]
       
    def test_sample(self):
        # use the SABC scheme for T = 1
        steps, epsilon, n_samples, n_samples_per_param = 1, 10, 10, 1
        sampler = SABC([self.model], [self.dist_calc], self.backend, seed = 1)
        journal = sampler.sample([self.observation], steps, epsilon, n_samples, n_samples_per_param)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[0]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))


        # use the SABC scheme for T = 2
        steps, epsilon, n_samples, n_samples_per_param = 2, 10, 10, 1
        sampler = SABC([self.model], [self.dist_calc], self.backend, seed = 1)
        journal = sampler.sample([self.observation], steps, epsilon, n_samples, n_samples_per_param)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))
        self.assertLess(mu_post_mean - 0.55859197, 10e-2)
        self.assertLess(sigma_post_mean - 7.03987723, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)

class ABCsubsimTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=0)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        #self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]
       
    def test_sample(self):

        # use the ABCsubsim scheme for T = 1
        steps, n_samples, n_samples_per_param = 1, 10, 1
        sampler = ABCsubsim([self.model], [self.dist_calc], self.backend, seed = 1)
        journal = sampler.sample([self.observation], steps, n_samples, n_samples_per_param)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))


        # use the ABCsubsim scheme for T = 2
        steps, n_samples, n_samples_per_param = 2, 10, 1
        sampler = ABCsubsim([self.model], [self.dist_calc], self.backend, seed = 1)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        journal = sampler.sample([self.observation], steps, n_samples, n_samples_per_param)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))
        self.assertLess(mu_post_mean - (-0.81410299), 10e-2)
        self.assertLess(sigma_post_mean - 9.25442675, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)


class SMCABCTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=0)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        #self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]

      
    def test_sample(self):
        # use the SMCABC scheme for T = 1
        steps, n_sample, n_simulate = 1, 10, 1
        sampler = SMCABC([self.model], [self.dist_calc], self.backend, seed = 1)
        journal = sampler.sample([self.observation], steps, n_sample, n_simulate)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))

        #self.assertEqual((mu_post_mean, sigma_post_mean), (,))
        
        # use the SMCABC scheme for T = 2
        T, n_sample, n_simulate = 2, 10, 1
        sampler = SMCABC([self.model], [self.dist_calc], self.backend, seed = 1)
        journal = sampler.sample([self.observation], T, n_sample, n_simulate)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))
        self.assertLess(mu_post_mean - (-0.786118677019), 10e-2)
        self.assertLess(sigma_post_mean - 4.63324738665, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)

class APMCABCTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=0)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        #self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]

      
    def test_sample(self):
        # use the APMCABC scheme for T = 1
        steps, n_sample, n_simulate = 1, 10, 1
        sampler = APMCABC([self.model], [self.dist_calc], self.backend, seed = 1)
        journal = sampler.sample([self.observation], steps, n_sample, n_simulate, alpha=.9)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))

        self.assertFalse(journal.number_of_simulations == 0)

        
        # use the APMCABC scheme for T = 2
        T, n_sample, n_simulate = 2, 10, 1
        sampler = APMCABC([self.model], [self.dist_calc], self.backend, seed = 1)
        journal = sampler.sample([self.observation], T, n_sample, n_simulate, alpha=.9)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))
        self.assertLess(mu_post_mean - (-3.397848324005792), 10e-2)
        self.assertLess(sigma_post_mean - 6.451434816944525, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)

class RSMCABCTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=0)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        #self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]

      
    def test_sample(self):
        # use the RSMCABC scheme for T = 1
        steps, n_sample, n_simulate = 1, 10, 1
        sampler = RSMCABC([self.model], [self.dist_calc], self.backend, seed = 1)
        journal = sampler.sample([self.observation], steps, n_sample, n_simulate)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))

        self.assertFalse(journal.number_of_simulations == 0)

        #self.assertEqual((mu_post_mean, sigma_post_mean), (,))
        
        # use the RSMCABC scheme for T = 2
        steps, n_sample, n_simulate = 2, 10, 1
        sampler = RSMCABC([self.model], [self.dist_calc], self.backend, seed = 1)
        journal = sampler.sample([self.observation], steps, n_sample, n_simulate)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                            (len(sigma_post_sample), sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10,1))
        self.assertEqual(sigma_sample_shape, (10,1))
        self.assertEqual(weights_sample_shape, (10,1))
        self.assertLess(mu_post_mean - (1.52651600439), 10e-2)
        self.assertLess(sigma_post_mean - 6.49994754262, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)

if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod
import numpy as np

try:
    import torch
except ImportError:
    has_torch = False
else:
    has_torch = True
    from abcpy.NN_utilities.utilities import load_net
    from abcpy.NN_utilities.networks import createDefaultNN


class Statistics(metaclass=ABCMeta):
    """This abstract base class defines how to calculate statistics from dataset.

    The base class also implements a polynomial expansion with cross-product
    terms that can be used to get desired polynomial expansion of the calculated statistics.


    """

    @abstractmethod
    def __init__(self, degree=1, cross=False, previous_statistics=None):
        """Constructor that must be overwritten by the sub-class.

        The constructor of a sub-class must accept arguments for the polynomial
        expansion after extraction of the summary statistics, one has to define
        the degree of polynomial expansion and cross, indicating whether cross-prodcut
        terms are included.

        Parameters
        ----------
        degree: integer, optional
            Of polynomial expansion. The default value is 2 meaning second order polynomial expansion.
        cross: boolean, optional
            Defines whether to include the cross-product terms. The default value is True, meaning the cross product term
            is included.
        previous_statistics: Statistics class, optional
            It allows pipelining of Statistics. Specifically, if the final statistic to be used is determined by the
            composition of two Statistics, you can pass the first here; then, whenever the final statistic is needed, it
            is sufficient to call the `statistics` method of the second one, and that will automatically apply both
            transformations.
        """

        raise NotImplementedError

    @abstractmethod
    def statistics(self, data: object) -> object:
        """To be overwritten by any sub-class: should extract statistics from the
        data set data. It is assumed that data is a  list of n same type
        elements(eg., The data can be a list containing n timeseries, n graphs or n np.ndarray).

        Parameters
        ----------
        data: python list
            Contains n data sets with length p.
        Returns
        -------
        numpy.ndarray
            nxp matrix where for each of the n data points p statistics are calculated.

        """

        raise NotImplementedError

    def _polynomial_expansion(self, summary_statistics):
        """Helper function that does the polynomial expansion and includes cross-product
        terms of summary_statistics, already calculated summary statistics.

        Parameters
        ----------
        summary_statistics: numpy.ndarray
            nxp matrix where n is number of data points in the datasets data set and p number os
            summary statistics calculated.
        Returns
        -------
        numpy.ndarray
            nx(p+degree*p+cross*nchoosek(p,2)) matrix where for each of the n pointss with
            p statistics, degree*p polynomial expansion term and cross*nchoosek(p,2) many
            cross-product terms are calculated.

        """

        # Check summary_statistics is a np.ndarry
        if not isinstance(summary_statistics, np.ndarray):
            raise TypeError('Summary statistics is not of allowed types')
        # Include the polynomial expansion
        result = summary_statistics
        for ind in range(2, self.degree + 1):
            result = np.column_stack((result, np.power(summary_statistics, ind)))

        # Include the cross-product term
        if self.cross == True and summary_statistics.shape[1] > 1:
            # Convert to a matrix
            for ind1 in range(0, summary_statistics.shape[1]):
                for ind2 in range(ind1 + 1, summary_statistics.shape[1]):
                    result = np.column_stack((result, summary_statistics[:, ind1] * summary_statistics[:, ind2]))
        return result

    def _check_and_transform_input(self, data):
        """
        """
        if isinstance(data, list):
            if np.array(data).shape == (len(data),):
                if len(data) == 1:
                    data = np.array(data).reshape(1, 1)
                data = np.array(data).reshape(len(data), 1)
            else:
                data = np.concatenate(data).reshape(len(data), -1)
        else:
            raise TypeError('Input data should be of type list, but found type {}'.format(type(data)))

        return data


class Identity(Statistics):
    """
    This class implements identity statistics not applying any transformation to the data, before the optional
    polynomial expansion step. If the data set contains n numpy.ndarray of length p, it returns therefore an
    nx(p+degree*p+cross*nchoosek(p,2)) matrix, where for each of the n points with p statistics, degree*p polynomial
    expansion term and cross*nchoosek(p,2) many cross-product terms are calculated.
    """

    def __init__(self, degree=1, cross=False, previous_statistics=None):
        """

        Parameters
        ----------
        degree : integer, optional
            Of polynomial expansion. The default value is 2 meaning second order polynomial expansion.
        cross : boolean, optional
            Defines whether to include the cross-product terms. The default value is True, meaning the cross product term
            is included.
        previous_statistics : Statistics class, optional
            It allows pipelining of Statistics. Specifically, if the final statistic to be used is determined by the
            composition of two Statistics, you can pass the first here; then, whenever the final statistic is needed, it
            is sufficient to call the `statistics` method of the second one, and that will automatically apply both
            transformations.
        """
        self.degree = degree
        self.cross = cross
        self.previous_statistics = previous_statistics

    def statistics(self, data):
        """
        Parameters
        ----------
        data: python list
            Contains n data sets with length p.
        Returns
        -------
        numpy.ndarray
            nx(p+degree*p+cross*nchoosek(p,2)) matrix where for each of the n data points with length p,
            (p+degree*p+cross*nchoosek(p,2)) statistics are calculated.
        """

        # pipeline: first call the previous statistics:
        if self.previous_statistics is not None:
            data = self.previous_statistics.statistics(data)
        # the first of the statistics need to take list as input, in order to match the API. Then actually the
        # transformations work on np.arrays. In fact the first statistic transforms the list to array. Therefore, the
        # following code needs to be called only if the self statistic is the first, i.e. it does not have a
        # previous_statistic element.
        else:
            data = self._check_and_transform_input(data)

        # Expand the data with polynomial expansion
        result = self._polynomial_expansion(data)

        return result


class LinearTransformation(Statistics):
    """Applies a linear transformation to the data to get (usually) a lower dimensional statistics. Then you can apply
    an additional polynomial expansion step.
    """

    def __init__(self, coefficients, degree=1, cross=False, previous_statistics=None):
        """
        Parameters
        ----------
        coefficients: coefficients is a matrix with size d x p, where d is the dimension of the summary statistic that
            is obtained after applying the linear transformation (i.e. before a possible polynomial expansion is
            applied), while d is the dimension of each data.
        degree : integer, optional
            Of polynomial expansion. The default value is 2 meaning second order polynomial expansion.
        cross : boolean, optional
            Defines whether to include the cross-product terms. The default value is True, meaning the cross product term
            is included.
        previous_statistics : Statistics class, optional
            It allows pipelining of Statistics. Specifically, if the final statistic to be used is determined by the
            composition of two Statistics, you can pass the first here; then, whenever the final statistic is needed, it
            is sufficient to call the `statistics` method of the second one, and that will automatically apply both
            transformations.
        """
        self.coefficients = coefficients
        self.degree = degree
        self.cross = cross
        self.previous_statistics = previous_statistics

    def statistics(self, data):
        """
        Parameters
        ----------
        data: python list
            Contains n data sets with length p.
        Returns
        -------
        numpy.ndarray
            nx(d+degree*d+cross*nchoosek(d,2)) matrix where for each of the n data points with length p you apply the
            linear transformation to get to dimension d, from where (d+degree*d+cross*nchoosek(d,2)) statistics are
            calculated.
        """

        # pipeline: first call the previous statistics:
        if self.previous_statistics is not None:
            data = self.previous_statistics.statistics(data)
        # the first of the statistics need to take list as input, in order to match the API. Then actually the
        # transformations work on np.arrays. In fact the first statistic transforms the list to array. Therefore, the
        # following code needs to be called only if the self statistic is the first, i.e. it does not have a
        # previous_statistic element.
        else:
            data = self._check_and_transform_input(data)

        # Apply now the linear transformation
        if not data.shape[1] == self.coefficients.shape[0]:
            raise ValueError('Mismatch in dimension of summary statistics and coefficients')
        result = np.dot(data, self.coefficients)

        # Expand the data with polynomial expansion
        result = self._polynomial_expansion(result)

        return result


class NeuralEmbedding(Statistics):
    """Computes the statistics by applying a neural network transformation. 
    
    It is essentially a wrapper for the application of a neural network transformation to the data. Note that the
    neural network has had to be trained in some way (for instance check the statistics learning routines) and that 
    Pytorch is required for this part to work.   
    """

    def __init__(self, net, previous_statistics=None):  # are these default values OK?
        """
        Parameters
        ----------
        net : torch.nn object
            the embedding neural network. The input size of the neural network must coincide with the size of each of
            the datapoints.
        previous_statistics : Statistics class, optional
            It allows pipelining of Statistics. Specifically, if the final statistic to be used is determined by the
            composition of two Statistics, you can pass the first here; then, whenever the final statistic is needed, it
            is sufficient to call the `statistics` method of the second one, and that will automatically apply both
            transformations.
        """
        if not has_torch:
            raise ImportError(
                "Pytorch is required to instantiate an element of the {} class, in order to handle "
                "neural networks. Please install it. ".format(self.__class__.__name__))

        self.net = net
        self.previous_statistics = previous_statistics

    @classmethod
    def fromFile(cls, path_to_net_state_dict, network_class=None, input_size=None, output_size=None, hidden_sizes=None,
                 previous_statistics=None):
        """If the neural network state_dict was saved to the disk, this method can be used to instantiate a
        NeuralEmbedding object with that neural network.

        In order for the state_dict to be read correctly, the network class is needed. Therefore, we provide 2 options:
        1) the Pytorch neural network class can be passed (if the user defined it, for instance)
        2) if the neural network was defined by using the DefaultNN class in abcpy.NN_utilities.networks, you can
        provide arguments `input_size`, `output_size` and `hidden_sizes` (the latter is optional) that define
        the sizes of a fully connected network; then a DefaultNN is instantiated with those sizes. This can be used
        if for instance the neural network was trained using the utilities in abcpy.statisticslearning and you did
        not provide explicitly the neural network class there, but defined it through the sizes of the different layers.

        In both cases, note that the input size of the neural network must coincide with the size of each of the
        datapoints generated from the model (unless some other statistics are computed beforehand).

        Parameters
        ----------
        path_to_net_state_dict : basestring
            the path where the state-dict is saved
        network_class : torch.nn class, optional
            if the neural network class is known explicitly (for instance if the used defined it), then it has to be
             passed here. This must not be provided together with `input_size` or `output_size`.
        input_size : integer, optional
            if the neural network is an instance of abcpy.NN_utilities.networks.DefaultNN with some input and
            output size, then you should provide here the input size of the network. It has to be provided together with
            the corresponding output_size, and it must not be provided with `network_class`.
        output_size : integer, optional
            if the neural network is an instance of abcpy.NN_utilities.networks.DefaultNN with some input and
            output size, then you should provide here the output size of the network. It has to be provided together
            with the corresponding input_size, and it must not be provided with `network_class`.
        hidden_sizes : array-like, optional
            if the neural network is an instance of abcpy.NN_utilities.networks.DefaultNN with some input and
            output size, then you can provide here an array-like with the size of the hidden layers (for instance
            [5,7,5] denotes 3 hidden layers with correspondingly 5,7,5 neurons). In case this parameter is not provided,
            the hidden sizes are determined from the input and output sizes as determined in
            abcpy.NN_utilities.networks.DefaultNN. Note that this must not be provided together with `network_class`.
        previous_statistics : Statistics class, optional
            It allows pipelining of Statistics. Specifically, if the final statistic to be used is determined by the
            composition of two Statistics, you can pass the first here; then, whenever the final statistic is needed, it
            is sufficient to call the `statistics` method of the second one, and that will automatically apply both
            transformations. In this case, this is the statistics that has to be computed before the neural network
            transformation is applied.
        Returns
        -------
        abcpy.statistics.NeuralEmbedding
            the `NeuralEmbedding` object with the neural network obtained from the specified file.
        """
        if not has_torch:
            raise ImportError(
                "Pytorch is required to instantiate an element of the {} class, in order to handle "
                "neural networks. Please install it. ".format(cls.__name__))

        if network_class is None and (input_size is None or output_size is None):
            raise RuntimeError("You need to pass either network class or both input_size and output_size.")
        if network_class is not None and (input_size is not None or output_size is not None):
            raise RuntimeError("You can't pass together network_class and one of input_size, output_size")
        if network_class is not None and hidden_sizes is not None:
            raise RuntimeError("You passed hidden_sizes as an argument, but that may be passed only if you are passing "
                               "input_size and input_size as well, and you are not passing network_class.")

        if network_class is not None:  # user explicitly passed the NN class
            net = load_net(path_to_net_state_dict, network_class)
            statistic_object = cls(net, previous_statistics=previous_statistics)
        else:  # the user passed the input_size, output_size and (maybe) the hidden_sizes
            net = load_net(path_to_net_state_dict, createDefaultNN(input_size=input_size, output_size=output_size,
                                                                   hidden_sizes=hidden_sizes))
            statistic_object = cls(net, previous_statistics=previous_statistics)

        return statistic_object

    def statistics(self, data):
        """
        Parameters
        ----------
        data: python list
            Contains n data sets with length p.
        Returns
        -------
        numpy.ndarray
            the statistics computed by applying the neural network.
        """

        # pipeline: first call the previous statistics:
        if self.previous_statistics is not None:
            data = self.previous_statistics.statistics(data)
        # the first of the statistics need to take list as input, in order to match the API. Then actually the
        # transformations work on np.arrays. In fact the first statistic transforms the list to array. Therefore, the
        # following code needs to be called only if the self statistic is the first, i.e. it does not have a
        # previous_statistic element.
        else:
            data = self._check_and_transform_input(data)

        data = torch.from_numpy(data.astype("float32"))

        # move data to gpu if the net is on gpu
        if next(self.net.parameters()).is_cuda:
            data = data.cuda()

        # simply apply the network transformation.
        result = self.net(data).cpu().detach().numpy()

        return np.array(result)



import numpy as np
from abcpy.probabilisticmodels import Hyperparameter, ModelResultingFromOperation


class GraphTools():
    """This class implements all methods that will be called recursively on the graph structure."""

    def sample_from_prior(self, model=None, rng=np.random.RandomState()):
        """
        Samples values for all random variables of the model.
        Commonly used to sample new parameter values on the whole graph.

        Parameters
        ----------
        model: abcpy.ProbabilisticModel object
            The root model for which sample_from_prior should be called.
        rng: Random number generator
            Defines the random number generator to be used
        """
        if(model is None):
            model = self.model
        # If it was at some point not possible to sample (due to incompatible parameter values provided by the parents), we start from scratch
        while(not(self._sample_from_prior(model, rng=rng))):
            self._reset_flags(model)

        # At the end of the algorithm, are flags are reset such that new methods can act on the graph freely
        self._reset_flags(model)

    def _sample_from_prior(self, models, is_not_root=False, was_accepted=True, rng=np.random.RandomState()):
        """
        Recursive version of sample_from_prior. Commonly called from within sample_from_prior.

        Parameters
        ----------
        models: list of abcpy.ProbabilisticModel objects
            Defines the models for which, together with their parents, new parameters will be sampled
        is_root: boolean
            Whether the probabilistic models provided in models are root models.
        was_accepted: boolean
            Whether the sampled values for all previous/parent models were accepted.
        rng: Random number generator
            Defines the random number generator to be used

        Returns
        -------
        boolean:
            Whether it was possible to sample new values for all nodes of the graph.
        """

        # If it was so far possible to sample parameters for all nodes, the current node as well as its parents are sampled, using depth-first search
        if(was_accepted):
            for model in models:
                for parent in model.get_input_models():
                    if(not(parent.visited)):
                        parent.visited = True
                        was_accepted = self._sample_from_prior([parent], is_not_root = True, was_accepted=was_accepted, rng=rng)
                        if(not(was_accepted)):
                            return False

                if(is_not_root and not(model._forward_simulate_and_store_output(rng=rng))):
                    return False

                model.visited = True

        return was_accepted

    def _reset_flags(self, models=None):
        """
        Resets all flags that say that a probabilistic model has been updated. Commonly used after actions on the whole
        graph, to ensure that new actions can take place.

        Parameters
        ----------
        models: list of abcpy.ProbabilisticModel
            The models for which, together with their parents, the flags should be reset. If no value is provided, the
            root models are assumed to be the model of the inference method.
        """
        if not models:
            models = self.model

        # For each model, the flags of the parents get reset recursively.
        for model in models:
            for parent in model.get_input_models():
                self._reset_flags([parent])
            model.visited = False
            model.calculated_pdf = None

    def pdf_of_prior(self, models, parameters, mapping=None, is_root=True):
        """
        Calculates the joint probability density function of the prior of the specified models at the given parameter values.
        Commonly used to check whether new parameters are valid given the prior, as well as to calculate acceptance probabilities.

        Parameters
        ----------
        models: list of abcpy.ProbabilisticModel objects
            Defines the models for which the pdf of their prior should be evaluated
        parameters: python list
            The parameters at which the pdf should be evaluated
        mapping: list of tupels
            Defines the mapping of probabilistic models and index in a parameter list.
        is_root: boolean
            A flag specifying whether the provided models are the root models. This is to ensure that the pdf is calculated correctly.

        Returns
        -------
        list
            The resulting pdf,as well as the next index to be considered in the parameters list.
        """
        self.set_parameters(parameters)
        result = self._recursion_pdf_of_prior(models, parameters, mapping, is_root)
        return result

    def _recursion_pdf_of_prior(self, models, parameters, mapping=None, is_root=True):
        """
        Calculates the joint probability density function of the prior of the specified models at the given parameter values.
        Commonly used to check whether new parameters are valid given the prior, as well as to calculate acceptance probabilities.

        Parameters
        ----------
        models: list of abcpy.ProbabilisticModel objects
            Defines the models for which the pdf of their prior should be evaluated
        parameters: python list
            The parameters at which the pdf should be evaluated
        mapping: list of tupels
            Defines the mapping of probabilistic models and index in a parameter list.
        is_root: boolean
            A flag specifying whether the provided models are the root models. This is to ensure that the pdf is calculated correctly.

        Returns
        -------
        list
            The resulting pdf,as well as the next index to be considered in the parameters list.
        """
        # At the beginning of calculation, obtain the mapping
        if(is_root):
            mapping, garbage_index = self._get_mapping()

        # The pdf of each root model is first calculated seperately
        result = [1.]*len(models)

        for i, model in enumerate(models):
            # If the model is not a root model, the pdf of this model, given the prior, should be calculated
            if(not(is_root) and not(isinstance(model, ModelResultingFromOperation))):
                # Define a helper list which will contain the parameters relevant to the current model for pdf calculation
                relevant_parameters = []

                for mapped_model, model_index in mapping:
                    if(mapped_model==model):
                        parameter_index = model_index
                        #for j in range(model.get_output_dimension()):
                        relevant_parameters.append(parameters[parameter_index])
                        #parameter_index+=1
                        break
                if(len(relevant_parameters)==1):
                    relevant_parameters = relevant_parameters[0]
                else:
                    relevant_parameters = np.array(relevant_parameters)
            else:
                relevant_parameters=[]

            # Mark whether the parents of each model have been visited before for this model to avoid repeated calculation.
            visited_parents = [False for j in range(len(model.get_input_models()))]
            # For each parent, the pdf of this parent has to be calculated as well.
            for parent_index, parent in enumerate(model.get_input_models()):
                # Only calculate the pdf if the parent has never been visited for this model
                if(not(visited_parents[parent_index])):
                    pdf = self._recursion_pdf_of_prior([parent], parameters, mapping=mapping, is_root=False)
                    input_models = model.get_input_models()
                    for j in range(len(input_models)):
                        if input_models[j][0] == parent:
                            visited_parents[j]=True
                    result[i] *= pdf
            if(not(is_root)):
                if(model.calculated_pdf is None):
                    result[i] *= model.pdf(model.get_input_values(),relevant_parameters)
                else:
                    result[i] *= 1 

        # Multiply the pdfs of all roots together to give an overall pdf.
        temporary_result = result
        result = 1.
        for individual_result in temporary_result:
            result *= individual_result

        return result

    def _get_mapping(self, models=None, index=0, is_not_root=False):
        """Returns a mapping of model and first index corresponding to the outputs in this model in parameter lists.

        Parameters
        ----------
        models: list
            List of abcpy.ProbabilisticModel objects
        index: integer
            Next index to be mapped in a parameter list
        is_not_root: boolean
            Specifies whether the models specified are root models.

        Returns
        -------
        list
            A list containing two entries. The first entry corresponds to the mapping of the root models, including their parents. The second entry corresponds to the next index to be considered in a parameter list.
        """

        if(models is None):
            models = self.model

        mapping = []

        for model in models:
            # If this model corresponds to an unvisited free parameter, add it to the mapping
            if(is_not_root and not(model.visited) and not(isinstance(model, Hyperparameter)) and not(isinstance(model, ModelResultingFromOperation))):
                mapping.append((model, index))
                index+= 1 #model.get_output_dimension()
            # Add all parents to the mapping, if applicable
            for parent in model.get_input_models():
                parent_mapping, index = self._get_mapping([parent], index=index, is_not_root=True)
                parent.visited=True
                for mappings in parent_mapping:
                    mapping.append(mappings)

            model.visited=True

        # At the end of the algorithm, reset all flags such that another method can act on the graph freely.
        if(not(is_not_root)):
            self._reset_flags()

        return [mapping, index]

    def _get_names_and_parameters(self):
        """
        A function returning the name of each model and the corresponding parameters to this model

        Returns
        -------
        list:
            Each entry is a tupel, the first entry of which is the name of the model and the second entry is the parameter values associated with it
        """
        mapping = self._get_mapping()[0]

        return_value = []

        for model, index in mapping:

            return_value.append((model.name, self.accepted_parameters_manager.get_accepted_parameters_bds_values([model])))

        return return_value


    def get_parameters(self, models=None, is_root=True):
        """
        Returns the current values of all free parameters in the model. Commonly used before perturbing the parameters
        of the model.

        Parameters
        ----------
        models: list of abcpy.ProbabilisticModel objects
            The models for which, together with their parents, the parameter values should be returned. If no value is
            provided, the root models are assumed to be the model of the inference method.
        is_root: boolean
            Specifies whether the current models are at the root. This ensures that the values corresponding to
            simulated observations will not be returned.

        Returns
        -------
        list
            A list containing all currently sampled values of the free parameters.
        """
        parameters = []

        # If we are at the root, we set models to the model attribute of the inference method
        if is_root:
            models = self.model

        for model in models:
            # If we are not at the root, the sampled values for the current node should be returned
            if is_root == False and not isinstance(model, (ModelResultingFromOperation, Hyperparameter)):
                parameters.append(model.get_stored_output_values())
                model.visited = True

            # Implement a depth-first search to return also the sampled values associated with each parent of the model
            for parent in model.get_input_models():
                if not parent.visited:
                    parameters += self.get_parameters(models=[parent], is_root=False)
                    parent.visited = True

        # At the end of the algorithm, are flags are reset such that new methods can act on the graph freely
        if is_root:
            self._reset_flags()

        return parameters


    def set_parameters(self, parameters, models=None, index=0, is_root=True):
        """
        Sets new values for the currently used values of each random variable.
        Commonly used after perturbing the parameter values using a kernel.

        Parameters
        ----------
        parameters: list
            Defines the values to which the respective parameter values of the models should be set
        model: list of abcpy.ProbabilisticModel objects
             Defines all models for which, together with their parents, new values should be set. If no value is provided, the root models are assumed to be the model of the inference method.
        index: integer
            The current index to be considered in the parameters list
        is_root: boolean
            Defines whether the current models are at the root. This ensures that only values corresponding to random variables will be set.

        Returns
        -------
        list: [boolean, integer]
            Returns whether it was possible to set all parameters and the next index to be considered in the parameters list.
        """

        # If we are at the root, we set models to the model attribute of the inference method
        if is_root:
            models = self.model

        for model in models:
            # New parameters should only be set in case we are not at the root
            if not is_root and not isinstance(model, ModelResultingFromOperation):
                #new_output_values = np.array(parameters[index:index + model.get_output_dimension()])
                new_output_values = np.array(parameters[index]).reshape(-1,)
                if not model.set_output_values(new_output_values):
                    return [False, index]
                index += 1 #model.get_output_dimension()
                model.visited = True

            # New parameters for all parents are set using a depth-first search
            for parent in model.get_input_models():
                if not parent.visited and not isinstance(parent, Hyperparameter):
                    is_set, index = self.set_parameters(parameters, models=[parent], index=index, is_root=False)
                    if not is_set:
                        # At the end of the algorithm, are flags are reset such that new methods can act on the graph freely
                        if is_root:
                            self._reset_flags()
                        return [False, index]
            model.visited = True

        # At the end of the algorithm, are flags are reset such that new methods can act on the graph freely
        if(is_root):
            self._reset_flags()

        return [True, index]

    def get_correct_ordering(self, parameters_and_models, models=None, is_root = True):
        """
        Orders the parameters returned by a kernel in the order required by the graph.
        Commonly used when perturbing the parameters.

        Parameters
        ----------
        parameters_and_models: list of tuples
            Contains tuples containing as the first entry the probabilistic model to be considered and as the second entry the parameter values associated with this model
        models: list
            Contains the root probabilistic models that make up the graph. If no value is provided, the root models are assumed to be the model of the inference method.

        Returns
        -------
        list
            The ordering which can be used by recursive functions on the graph.
        """
        ordered_parameters = []

        # If we are at the root, we set models to the model attribute of the inference method
        if(is_root):
            models=self.model

        for model in models:
            if(not(model.visited)):
                model.visited = True

                # Check all entries in parameters_and_models to determine whether the current model is contained within it
                for corresponding_model, parameter in parameters_and_models:
                    if(corresponding_model==model):
                        for param in parameter:
                            ordered_parameters.append(param)
                        break

                # Recursively order all the parents of the current model
                for parent in model.get_input_models():
                    if(not(parent.visited)):
                        parent_ordering = self.get_correct_ordering(parameters_and_models, models=[parent],is_root=False)
                        for parent_parameters in parent_ordering:
                            ordered_parameters.append(parent_parameters)

        # At the end of the algorithm, are flags are reset such that new methods can act on the graph freely
        if(is_root):
            self._reset_flags()

        return ordered_parameters

    def simulate(self, n_samples_per_param, rng=np.random.RandomState(), npc=None):
        """Simulates data of each model using the currently sampled or perturbed parameters.

        Parameters
        ----------
        rng: random number generator
            The random number generator to be used.

        Returns
        -------
        list
            Each entry corresponds to the simulated data of one model.
        """
        result = []
        for model in self.model:
            parameters_compatible = model._check_input(model.get_input_values())
            if parameters_compatible:
                if npc is not None and npc.communicator().Get_size() > 1:
                    simulation_result = npc.run_nested(model.forward_simulate, model.get_input_values(), n_samples_per_param, rng=rng)
                else:
                    simulation_result = model.forward_simulate(model.get_input_values(),n_samples_per_param, rng=rng)
                result.append(simulation_result)
            else:
                return None
        return result



import unittest
import numpy as np

from abcpy.approx_lhd import SynLikelihood
from abcpy.statistics import Identity
from abcpy.continuousmodels import Normal, Uniform
from abcpy.jointapprox_lhd import ProductCombination

class ProductCombinationTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc1 = Identity(degree = 1, cross = 0)
        self.stat_calc2 = Identity(degree= 1, cross = 0)
        self.likfun1 = SynLikelihood(self.stat_calc1)
        self.likfun2 = SynLikelihood(self.stat_calc2)
        ## Define Models
        # define a uniform prior distribution
        self.mu = Uniform([[-5.0], [5.0]], name='mu')
        self.sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model1 = Normal([self.mu,self.sigma])
        self.model2 = Normal([self.mu,self.sigma])

        #Check whether wrong sized distnacefuncs gives an error
        self.assertRaises(ValueError, ProductCombination, [self.model1,self.model2], [self.likfun1])

        self.jointapprox_lhd = ProductCombination([self.model1, self.model2], [self.likfun1, self.likfun2])

    def test_likelihood(self):
        # test simple distance computation
        a = [[0, 0, 0],[0, 0, 0]]
        b = [[0, 0, 0],[0, 0, 0]]
        c = [[1, 1, 1],[1, 1, 1]]

        #Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.jointapprox_lhd.likelihood, 3.4, [[2,1]])
        self.assertRaises(TypeError, self.jointapprox_lhd.likelihood, [[2,4]], 3.4)

        # test input has different dimensionality
        self.assertRaises(BaseException, self.jointapprox_lhd.likelihood, [a], [b,c])
        self.assertRaises(BaseException, self.jointapprox_lhd.likelihood, [b,c], [a])

        # test whether they compute correct values
        # create observed data
        y_obs = [[9.8], [9.8]]
        # create fake simulated data
        self.mu._fixed_values = [1.1]
        self.sigma._fixed_values = [1.0]
        y_sim_1 = self.model1.forward_simulate(self.model1.get_input_values(), 100, rng=np.random.RandomState(1))
        y_sim_2 = self.model2.forward_simulate(self.model2.get_input_values(), 100, rng=np.random.RandomState(1))
        # calculate the statistics of the observed data
        comp_likelihood = self.jointapprox_lhd.likelihood(y_obs, [y_sim_1, y_sim_2])
        expected_likelihood = 8.612491843767518e-43
        # This checks whether it computes a correct value and dimension is right
        self.assertLess(comp_likelihood - expected_likelihood, 10e-2)


if __name__ == '__main__':
    unittest.main()




from abcpy.backends import Backend, PDS, BDS

class BackendSpark(Backend):
    """
    A parallelization backend for Apache Spark. It is essetially a wrapper for
    the required Spark functionality.
    """
    
    def __init__(self, sparkContext, parallelism=4):
        """
        Initialize the backend with an existing and configured SparkContext.

        Parameters
        ----------
        sparkContext: pyspark.SparkContext
            an existing and fully configured PySpark context
        parallelism: int
            defines on how many workers a distributed dataset can be distributed
        """
        self.sc = sparkContext
        self.parallelism = parallelism


    def parallelize(self, python_list):
        """
        This is a wrapper of pyspark.SparkContext.parallelize().

        Parameters
        ----------
        list: Python list
            list that is distributed on the workers
        
        Returns
        -------
        PDSSpark class (parallel data set)
            A reference object that represents the parallelized list
        """
        
        rdd = self.sc.parallelize(python_list, self.parallelism)
        pds = PDSSpark(rdd)
        return pds


    def broadcast(self, object):
        """
        This is a wrapper for pyspark.SparkContext.broadcast().

        Parameters
        ----------
        object: Python object
            An abitrary object that should be available on all workers
        Returns
        -------
        BDSSpark class (broadcast data set)
            A reference to the broadcasted object
        """
        
        bcv = self.sc.broadcast(object)
        bds = BDSSpark(bcv)
        return bds


    def map(self, func, pds):
        """
        This is a wrapper for pyspark.rdd.map()

        Parameters
        ----------
        func: Python func
            A function that can be applied to every element of the pds
        pds: PDSSpark class
            A parallel data set to which func should be applied
        Returns
        -------
        PDSSpark class
            a new parallel data set that contains the result of the map
        """
        
        rdd = pds.rdd.map(func)
        new_pds = PDSSpark(rdd)
        return new_pds


    def collect(self, pds):
        """
        A wrapper for pyspark.rdd.collect()

        Parameters
        ----------
        pds: PDSSpark class
            a parallel data set
        Returns
        -------
        Python list
            all elements of pds as a list
        """
        
        python_list = pds.rdd.collect()
        return python_list

    
    
class PDSSpark(PDS):
    """
    This is a wrapper for Apache Spark RDDs.
    """
    
    def __init__(self, rdd):
        """
        Returns
        -------
        rdd: pyspark.rdd
            initialize with an Spark RDD
        """
        
        self.rdd = rdd



class BDSSpark(BDS):
    """
    This is a wrapper for Apache Spark Broadcast variables.
    """
    
    def __init__(self, bcv):
        """
        Parameters
        ----------
        bcv: pyspark.broadcast.Broadcast
            Initialize with a Spark broadcast variable
        """
        
        self.bcv = bcv


    def value(self):
        """
        Returns
        -------
        object
            returns the referenced object that was broadcasted.
        """
        
        return self.bcv.value



from tqdm import tqdm
import logging


def fit(train_loader, model, loss_fn, optimizer, scheduler, n_epochs, cuda, start_epoch=0):
    """
    Basic function to train a neural network given a train_loader, a loss function and an optimizer.

    Loaders, model, loss function and metrics should work together for a given task,
    i.e. The model should be able to process data output of loaders,
    loss function should process target output of loaders and outputs from the model

    Examples: Classification: batch loader, classification model, NLL loss, accuracy metric
    Siamese network: Siamese loader, siamese model, contrastive loss

    Adapted from https://github.com/adambielski/siamese-triplet
    """

    logger = logging.getLogger("NN Trainer")

    for epoch in range(0, start_epoch):
        scheduler.step()

    for epoch in tqdm(range(start_epoch, n_epochs)):
        scheduler.step()

        # Train stage
        train_loss = train_epoch(train_loader, model, loss_fn, optimizer, cuda)

        logger.debug('Epoch: {}/{}. Train set: Average loss: {:.4f}'.format(epoch + 1, n_epochs, train_loss))


def train_epoch(train_loader, model, loss_fn, optimizer, cuda):
    """Function implementing the training in one epoch.

    Adapted from https://github.com/adambielski/siamese-triplet
    """
    model.train()
    losses = []
    total_loss = 0

    for batch_idx, (data, target) in enumerate(train_loader):
        target = target if len(target) > 0 else None
        if not type(data) in (tuple, list):
            data = (data,)
        if cuda:
            data = tuple(d.cuda() for d in data)
            if target is not None:
                target = target.cuda()

        optimizer.zero_grad()
        outputs = model(*data)

        if type(outputs) not in (tuple, list):
            outputs = (outputs,)

        loss_inputs = outputs
        if target is not None:
            target = (target,)
            loss_inputs += target

        loss_outputs = loss_fn(*loss_inputs)
        loss = loss_outputs[0] if type(loss_outputs) in (tuple, list) else loss_outputs
        losses.append(loss.item())
        total_loss += loss.item()
        loss.backward()
        optimizer.step()

        losses = []

    total_loss /= (batch_idx + 1)
    return total_loss



import unittest
from abcpy.inferences import *
from abcpy.continuousmodels import *
from abcpy.discretemodels import *
from abcpy.distances import LogReg
from abcpy.statistics import Identity
from abcpy.backends import BackendDummy as Backend
from abcpy.perturbationkernel import *

"""Tests whether the methods defined for operations on the graph work as intended."""


class SampleFromPriorTests(unittest.TestCase):
    """Tests whether sample_from_prior assigns new values to all nodes corresponding to free parameters in the graph."""
    def test(self):
        B1 = Binomial([10, 0.2])
        N1 = Normal([0.03, 0.01])
        N2 = Normal([0.1, N1])
        graph = Normal([B1, N2])

        statistics_calculator = Identity(degree = 2, cross = False)
        distance_calculator = LogReg(statistics_calculator)
        backend = Backend()

        sampler = RejectionABC([graph], [distance_calculator], backend)

        rng = np.random.RandomState(1)

        sampler.sample_from_prior(rng=rng)
        self.assertIsNotNone(B1._fixed_values)
        self.assertIsNotNone(N1._fixed_values)
        self.assertIsNotNone(N2._fixed_values)


class ResetFlagsTests(unittest.TestCase):
    """Tests whether it is possible to reset all visited flags in the graph."""
    def test(self):
        N1 = Normal([1, 0.1])
        N2 = Normal([N1, 0.1])
        N2.visited = True
        N1.visited = True

        statistics_calculator = Identity(degree=2, cross=False)
        distance_calculator = LogReg(statistics_calculator)
        backend = Backend()

        sampler = RejectionABC([N2], [distance_calculator], backend)

        sampler._reset_flags()

        self.assertFalse(N1.visited)
        self.assertFalse(N2.visited)


class GetParametersTests(unittest.TestCase):
    """Tests whether get_stored_output_values returns only the free parameters of the graph."""
    def setUp(self):
        self.B1 = Binomial([10, 0.2])
        self.N1 = Normal([0.03, 0.01])
        self.N2 = Normal([0.1, self.N1])
        self.graph = Normal([self.B1, self.N2])

        statistics_calculator = Identity(degree=2, cross=False)
        distance_calculator = LogReg(statistics_calculator)
        backend = Backend()

        self.sampler = RejectionABC([self.graph], [distance_calculator], backend)

        self.rng = np.random.RandomState(1)

        self.sampler.sample_from_prior(rng=self.rng)

    def test(self):
        free_parameters = self.sampler.get_parameters()
        self.assertEqual(len(free_parameters),3)


class SetParametersTests(unittest.TestCase):
    """Tests whether it is possible to set values for all free parameters of the graph."""
    def setUp(self):
        self.B1 = Binomial([10, 0.2])
        self.N1 = Normal([0.03, 0.01])
        self.N2 = Normal([0.1, self.N1])
        self.graph = Normal([self.B1, self.N2])

        statistics_calculator = Identity(degree=2, cross=False)
        distance_calculator = LogReg(statistics_calculator)
        backend = Backend()

        self.sampler = RejectionABC([self.graph], [distance_calculator], backend)

        self.rng = np.random.RandomState(1)

        self.sampler.sample_from_prior(rng=self.rng)

    def test(self):
        is_accepted, index = self.sampler.set_parameters([3, 0.12, 0.029])
        self.assertTrue(is_accepted)

        self.assertEqual(self.B1.get_stored_output_values()[0], 3)
        self.assertEqual(self.N1.get_stored_output_values()[0], 0.029)
        self.assertEqual(self.N2.get_stored_output_values()[0], 0.12)


class GetCorrectOrderingTests(unittest.TestCase):
    """Tests whether get_correct_ordering will order the values of free parameters in recursive dfs order."""
    def setUp(self):
        self.B1 = Binomial([10, 0.2])
        self.N1 = Normal([0.03, 0.01])
        self.N2 = Normal([0.1, self.N1])
        self.graph = Normal([self.B1, self.N2])

        statistics_calculator = Identity(degree=2, cross=False)
        distance_calculator = LogReg(statistics_calculator)
        backend = Backend()

        self.sampler = RejectionABC([self.graph], [distance_calculator], backend)

        self.rng = np.random.RandomState(1)

        self.sampler.sample_from_prior(rng=self.rng)

    def test(self):
        parameters_and_models = [(self.N1, [0.029]), (self.B1, [3]), (self.N2, [0.12])]
        ordered_parameters = self.sampler.get_correct_ordering(parameters_and_models)
        self.assertEqual(ordered_parameters, [3,0.12,0.029])


class PerturbTests(unittest.TestCase):
    """Tests whether perturb will change all fixed values for free parameters."""
    def setUp(self):
        self.B1 = Binomial([10, 0.2])
        self.N1 = Normal([0.03, 0.01])
        self.N2 = Normal([0.1, self.N1])
        self.graph = Normal([self.B1, self.N2])

        statistics_calculator = Identity(degree=2, cross=False)
        distance_calculator = LogReg(statistics_calculator)
        backend = Backend()

        self.sampler = PMCABC([self.graph], [distance_calculator], backend)

        self.rng = np.random.RandomState(1)

        self.sampler.sample_from_prior(rng=self.rng)

        kernel = DefaultKernel([self.N1, self.N2, self.B1])
        self.sampler.kernel = kernel

        self.sampler.accepted_parameters_manager.update_broadcast(self.sampler.backend, [[3, 0.11, 0.029],[4,0.098, 0.031]], accepted_cov_mats = [[[1,0],[0,1]]], accepted_weights=np.array([1,1]))


        kernel_parameters = []
        for kernel in self.sampler.kernel.kernels:
            kernel_parameters.append(
                self.sampler.accepted_parameters_manager.get_accepted_parameters_bds_values(kernel.models))
        self.sampler.accepted_parameters_manager.update_kernel_values(self.sampler.backend, kernel_parameters)


    def test(self):
        B1_value = self.B1.get_stored_output_values()
        N1_value = self.N1.get_stored_output_values()
        N2_value = self.N2.get_stored_output_values()

        self.sampler.perturb(1, rng=self.rng)

        self.assertNotEqual(B1_value, self.B1._fixed_values)
        self.assertNotEqual(N1_value, self.N1._fixed_values)
        self.assertNotEqual(N2_value, self.N2._fixed_values)


class SimulateTests(unittest.TestCase):
    """Tests whether the simulated data for multiple models has the correct format."""
    def test(self):
        B1 = Binomial([10, 0.2])
        N1 = Normal([0.03, 0.01])
        N2 = Normal([0.1, N1])
        graph1 = Normal([B1, N2])
        graph2 = Normal([1, N2])

        statistics_calculator = Identity(degree=2, cross=False)
        distance_calculator = LogReg(statistics_calculator)
        backend = Backend()

        sampler = RejectionABC([graph1, graph2], [distance_calculator, distance_calculator], backend)

        rng = np.random.RandomState(1)

        sampler.sample_from_prior(rng=rng)

        y_sim = sampler.simulate(1, rng=rng)

        self.assertTrue(isinstance(y_sim, list))

        self.assertTrue(len(y_sim)==2)

        self.assertTrue(isinstance(y_sim[0][0], np.ndarray))


class GetMappingTests(unittest.TestCase):
    """Tests whether the private get_mapping method will return the correct mapping."""
    def test(self):
        B1 = Binomial([10, 0.2])
        N1 = Normal([0.03, 0.01])
        N2 = Normal([0.1, N1])
        graph1 = Normal([B1, N2])
        graph2 = Normal([1, N2])

        statistics_calculator = Identity(degree=2, cross=False)
        distance_calculator = LogReg(statistics_calculator)
        backend = Backend()

        sampler = RejectionABC([graph1, graph2], [distance_calculator, distance_calculator], backend)

        rng = np.random.RandomState(1)

        sampler.sample_from_prior(rng=rng)

        mapping, index = sampler._get_mapping()
        self.assertTrue(mapping==[(B1, 0),(N2, 1),(N1,2)])

from abcpy.continuousmodels import Uniform

class PdfOfPriorTests(unittest.TestCase):
    """Tests the implemetation of pdf_of_prior"""
    def setUp(self):
        class Mockobject(Normal):
            def __init__(self, parameters):
                super(Mockobject, self).__init__(parameters)
            def pdf(self, input_values, x):
                return x

        self.N1 = Uniform([[1.3], [1.55]], name='n1')
        self.N2 = Uniform([[self.N1], [1.60]], name='n2')
        self.N3 = Uniform([[2], [4]], name='n3')
        self.graph1 = Uniform([[self.N1], [self.N2]], name='graph1')
        self.graph2 = Uniform([[.5], [self.N3]])

        self.graph = [self.graph1, self.graph2]

        statistics_calculator = Identity(degree=2, cross=False)
        distance_calculator = LogReg(statistics_calculator)
        backend = Backend()

        self.sampler1 = RejectionABC([self.graph1], [distance_calculator], backend)
        self.sampler2 = RejectionABC([self.graph2], [distance_calculator], backend)
        self.sampler3 = RejectionABC(self.graph, [distance_calculator, distance_calculator], backend)

        self.pdf1 = self.sampler1.pdf_of_prior(self.sampler1.model,  [1.32088846, 1.42945274])
        self.pdf2 = self.sampler2.pdf_of_prior(self.sampler2.model, [3])
        self.pdf3 = self.sampler3.pdf_of_prior(self.sampler3.model, [1.32088846, 1.42945274, 3])

    def test_return_value(self):
        """Tests whether the return value is float."""
        self.assertTrue(isinstance(self.pdf1, float))
        self.assertTrue(isinstance(self.pdf2, float))
        self.assertTrue(isinstance(self.pdf3, float))

    def test_result(self):
        """Test whether pdf calculation works as intended"""
        self.assertTrue(self.pdf1 == 14.331188169432183)
        self.assertTrue(self.pdf2 == 0.5)
        self.assertTrue(self.pdf3 == 7.1655940847160915)


if __name__ == '__main__':
    unittest.main()






import unittest
from mpi4py import MPI
from abcpy.backends import BackendMPI,BackendMPITestHelper
import numpy

def setUpModule():
    '''
    If an exception is raised in a setUpModule then none of 
    the tests in the module will be run. 
    
    This is useful because the teams run in a while loop on initialization
    only responding to the scheduler's commands and will never execute anything else.

    On termination of scheduler, the teams call quit() that raises a SystemExit(). 
    Because of the behaviour of setUpModule, it will not run any unit tests
    for the team and we now only need to write unit-tests from the scheduler's 
    point of view. 
    '''
    global rank,backend_mpi
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    backend_mpi = BackendMPI(process_per_model=2)

class MPIBackendTests(unittest.TestCase):

    def test_parallelize(self):
        data = [0]*backend_mpi.size()
        pds = backend_mpi.parallelize(data)
        pds_map = backend_mpi.map(lambda x, npc=None: x + MPI.COMM_WORLD.Get_rank(), pds)
        res = backend_mpi.collect(pds_map)

        for scheduler_index in backend_mpi.scheduler_node_ranks():
            self.assertTrue(scheduler_index not in res,"Node in scheduler_node_ranks performed map.")

    def test_map(self):
        def square_mpi(x, npc=None):
            local_res = numpy.array([2*(x**2)], 'i')
            #global_res = numpy.array([0], 'i')
            #MPI.COMM_WORLD.Reduce([local_res,MPI.INT], [global_res,MPI.INT], op=MPI.SUM, root=0)
            return local_res[0]
        
        data = [1,2,3,4,5]
        pds = backend_mpi.parallelize(data)
        pds_map = backend_mpi.map(square_mpi, pds)
        res = backend_mpi.collect(pds_map)
        assert res==list(map(lambda x:2*(x**2),data))


    def test_broadcast(self):
        data = [1,2,3,4,5]
        pds = backend_mpi.parallelize(data)

        bds = backend_mpi.broadcast(100)

        #Pollute the BDS values of the scheduler to confirm teams
        # use their broadcasted value
        for k,v in  backend_mpi.bds_store.items():
             backend_mpi.bds_store[k] = 99999

        def test_map(x, npc=None):
            return x + bds.value()

        pds_m = backend_mpi.map(test_map, pds)
        self.assertTrue(backend_mpi.collect(pds_m)==[101,102,103,104,105])

    def test_pds_delete(self):

        def check_if_exists(x, npc):
            obj = BackendMPITestHelper()
            if npc.communicator().Get_rank() == 0:
                return obj.check_pds(x)
            return None

        data = [1,2,3,4,5]
        pds = backend_mpi.parallelize(data)

        #Check if the pds we just created exists in all the teams(+scheduler)

        id_check_pds = backend_mpi.parallelize([pds.pds_id]*5)
        pds_check_result = backend_mpi.map(check_if_exists, id_check_pds)
        self.assertTrue(False not in backend_mpi.collect(pds_check_result),"PDS was not created")

        #Delete the PDS on scheduler and try again
        del pds
        pds_check_result = backend_mpi.map(check_if_exists,id_check_pds)

        self.assertTrue(True not in backend_mpi.collect(pds_check_result),"PDS was not deleted")


    def test_bds_delete(self):
        
        def check_if_exists(x, npc=None):
            obj = BackendMPITestHelper()
            return obj.check_bds(x)

        data = [1,2,3,4,5]
        bds = backend_mpi.broadcast(data)

        #Check if the pds we just created exists in all the teams(+scheduler)
        id_check_bds = backend_mpi.parallelize([bds.bds_id]*5)
        bds_check_result = backend_mpi.map(check_if_exists, id_check_bds)
        self.assertTrue(False not in backend_mpi.collect(bds_check_result),"BDS was not created")

        #Delete the PDS on scheduler and try again
        del bds
        bds_check_result = backend_mpi.map(check_if_exists,id_check_bds)
        self.assertTrue(True not in backend_mpi.collect(bds_check_result),"BDS was not deleted")


    def test_function_pickle(self):

        def square_mpi(x, npc=None):
            local_res = numpy.array([2*(x**2)], 'i')
            #global_res = numpy.array([0], 'i')
            #model_comm.Reduce([local_res,MPI.INT], [global_res,MPI.INT], op=MPI.SUM, root=0)
            return local_res[0]

        class staticfunctest_mpi:
            @staticmethod 
            def square_mpi(x, npc=None):
                local_res = numpy.array([2*(x**2)], 'i')
                #global_res = numpy.array([0], 'i')
                #model_comm.Reduce([local_res,MPI.INT], [global_res,MPI.INT], op=MPI.SUM, root=0)
                return local_res[0]

        class nonstaticfunctest_mpi:
            def square_mpi(self, x, npc=None):
                local_res = numpy.array([2*(x**2)], 'i')
                #global_res = numpy.array([0], 'i')
                #model_comm.Reduce([local_res,MPI.INT], [global_res,MPI.INT], op=MPI.SUM, root=0)
                return local_res[0]

        data = [1,2,3,4,5]
        expected_result = [2,8,18,32,50]

        pds = backend_mpi.parallelize(data)
        pds_map1 = backend_mpi.map(square_mpi,pds)
        pds_res1 = backend_mpi.collect(pds_map1)
        
        self.assertTrue(pds_res1==expected_result,"Failed pickle test for general function")

        pds_map3 = backend_mpi.map(staticfunctest_mpi.square_mpi,pds)
        pds_res3 = backend_mpi.collect(pds_map3)
        self.assertTrue(pds_res3==expected_result,"Failed pickle test for static function")

        obj = nonstaticfunctest_mpi()
        pds_map4 = backend_mpi.map(obj.square_mpi ,pds)
        pds_res4 = backend_mpi.collect(pds_map4)
        self.assertTrue(pds_res4==expected_result,"Failed pickle test for non-static function")



import unittest
import numpy as np

from abcpy.distances import Euclidean, PenLogReg, LogReg
from abcpy.statistics import Identity

class EuclideanTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree = 1, cross = 0)
        self.distancefunc = Euclidean(self.stat_calc)
    
    def test_distance(self):
        # test simple distance computation
        a = [[0, 0, 0],[0, 0, 0]]
        b = [[0, 0, 0],[0, 0, 0]]
        c = [[1, 1, 1],[1, 1, 1]]
        #Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc.distance, 3.4, b)
        self.assertRaises(TypeError, self.distancefunc.distance, a, 3.4)

        # test input has different dimensionality
        self.assertRaises(BaseException, self.distancefunc.distance, a, np.array([[0, 0], [1, 2]]))  
        self.assertRaises(BaseException, self.distancefunc.distance, a, np.array([[0, 0, 0], [1, 2, 3], [4, 5, 6]]))        

        # test whether they compute correct values
        self.assertTrue(self.distancefunc.distance(a,b) == np.array([0]))
        self.assertTrue(self.distancefunc.distance(a,c) == np.array([1.7320508075688772]))
        
    def test_dist_max(self):
        self.assertTrue(self.distancefunc.dist_max() == np.inf)        


class PenLogRegTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree = 1, cross = 0)
        self.distancefunc = PenLogReg(self.stat_calc)
    
    def test_distance(self):
        d1 = 0.5 * np.random.randn(100,2) - 10
        d2 = 0.5 * np.random.randn(100,2) + 10

        d1=d1.tolist()
        d2=d2.tolist()
        #Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc.distance, d1, 3.4)

        # completely separable datasets should have a distance of 1.0
        self.assertEqual(self.distancefunc.distance(d1,d2), 1.0)

        # equal data sets should have a distance of 0.0
        self.assertEqual(self.distancefunc.distance(d1,d1), 0.0)
        
    def test_dist_max(self):
        self.assertTrue(self.distancefunc.dist_max() == 1.0)


class LogRegTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree = 1, cross = 0)
        self.distancefunc = LogReg(self.stat_calc)
        
    def test_distance(self):
        d1 = 0.5 * np.random.randn(100,2) - 10
        d2 = 0.5 * np.random.randn(100,2) + 10

        d1=d1.tolist()
        d2=d2.tolist()
        
        #Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc.distance, d1, 3.4)
        
        # completely separable datasets should have a distance of 1.0
        self.assertEqual(self.distancefunc.distance(d1,d2), 1.0)

        # equal data sets should have a distance of 0.0
        self.assertEqual(self.distancefunc.distance(d1,d1), 0.0)
        
    def test_dist_max(self):
        self.assertTrue(self.distancefunc.dist_max() == 1.0)

if __name__ == '__main__':
    unittest.main()



from abcpy.continuousmodels import *
from tests.probabilisticmodels_tests import AbstractAPIImplementationTests

import unittest

"""Tests whether the methods defined for continuous probabilistic models are working as intended."""


class UniformAPITests(AbstractAPIImplementationTests, unittest.TestCase):
    model_types = [Uniform]
    model_inputs = [[[0, 1], [1, 2]]]

class NormalAPITests(AbstractAPIImplementationTests, unittest.TestCase):
    model_types = [Normal]
    model_inputs = [[0,1]]

class StundentTAPITests(AbstractAPIImplementationTests, unittest.TestCase):
    model_types = [StudentT]
    model_inputs = [[0, 3]]

class MultivariateNormalAPITests(AbstractAPIImplementationTests, unittest.TestCase):
    model_types = [MultivariateNormal]
    model_inputs = [[[1, 0], [[1, 0], [0, 1]]]]

class MultiStudentTAPITests(AbstractAPIImplementationTests, unittest.TestCase):
    model_types = [MultiStudentT]
    model_inputs = [[[1, 0], [[1, 0], [0, 1]], 3]]


class CheckParametersAtInitializationTests(unittest.TestCase):
    """Tests that no probabilistic model with invalid parameters can be initialized."""

    # TODO: Test for all distributions the behaviour if input parameters are real distributions and not only
    # hyperparameters

    def test_Uniform(self):
        with self.assertRaises(ValueError):
            Uniform([[1, 2, 3], [4, 5]])

    def test_Normal(self):
        with self.assertRaises(ValueError):
            Normal([1, -0.1])

    def test_StudentT(self):
        with self.assertRaises(ValueError):
            StudentT([1, 0])

    def test_MultivariateNormal(self):
        with self.assertRaises(ValueError):
            MultivariateNormal([[1]])

        with self.assertRaises(ValueError):
            MultivariateNormal([[1, 0, 0], [[1, 0], [0, 1]]])

        with self.assertRaises(ValueError):
            MultivariateNormal([[1, 0], [[1, 1], [0, 1]]])

        with self.assertRaises(ValueError):
            MultivariateNormal([[1, 0], [[-1, 0], [0, -1]]])

    def test_MultiStudentT(self):
        with self.assertRaises(ValueError):
            MultiStudentT([[1, 0], [[1, 1], [0, 1]], 1])

        with self.assertRaises(ValueError):
            MultiStudentT([[1, 0], [[-1, 0], [0, -1]], 1])

        with self.assertRaises(ValueError):
            MultiStudentT([[1, 0], [[1, 0], [0, 1]], -1])



class DimensionTests(unittest.TestCase):
    """Tests whether the dimensions of all continuous models are defined in the correct way."""

    def test_Uniform(self):
        U = Uniform([[0, 1], [1, 2]])
        self.assertTrue(U.get_output_dimension()==2)

    def test_Normal(self):
        N = Normal([1, 0.1])
        self.assertTrue(N.get_output_dimension()==1)

    def test_StudentT(self):
        S = StudentT([3, 1])
        self.assertTrue(S.get_output_dimension()==1)

    def test_MultivariateNormal(self):
        M = MultivariateNormal([[1, 0], [[1, 0], [0, 1]]])
        self.assertTrue(M.get_output_dimension()==2)

    def test_MultiStudentT(self):
        M = MultiStudentT([[1, 0], [[0.1, 0], [0, 0.1]], 1])
        self.assertTrue(M.get_output_dimension()==2)



class SampleFromDistributionTests(unittest.TestCase):
    """Tests the return value of forward_simulate for all continuous distributions."""
    def test_Normal(self):
        N = Normal([1, 0.1])
        samples = N.forward_simulate(N.get_input_values(), 3)
        self.assertTrue(isinstance(samples, list))
        self.assertTrue(len(samples)==3)

    def test_MultivariateNormal(self):
        M = MultivariateNormal([[1, 0], [[0.1, 0], [0, 0.1]]])
        samples = M.forward_simulate(M.get_input_values(), 3)
        self.assertTrue(isinstance(samples, list))
        self.assertTrue(len(samples) == 3)

    def test_StudentT(self):
        S = StudentT([3, 1])
        samples = S.forward_simulate(S.get_input_values(), 3)
        self.assertTrue(isinstance(samples, list))
        self.assertTrue(len(samples) == 3)

    def test_MultiStudentT(self):
        S = MultiStudentT([[1, 0], [[0.1, 0], [0, 0.1]], 1])
        samples = S.forward_simulate(S.get_input_values(), 3)
        self.assertTrue(isinstance(samples, list))
        self.assertTrue(len(samples) == 3)

    def test_Uniform(self):
        U = Uniform([[0, 1], [1, 2]])
        samples = U.forward_simulate(U.get_input_values(), 3)
        self.assertTrue(isinstance(samples, list))
        self.assertTrue(len(samples) == 3)


class CheckParametersBeforeSamplingTests(unittest.TestCase):
    """Tests whether False will be returned if the input parameters of _check_parameters_before_sampling are not accepted."""

    def test_Uniform(self):
        U = Uniform([[0, 1], [1, 2]])
        self.assertFalse(U._check_input([1, 1, 0, 2]))
        self.assertFalse(U._check_input([1, 1, 2, 0]))

    def test_Normal(self):
        N = Normal([1, 0.1])
        self.assertFalse(N._check_input([1, -0.1]))

    def test_StudentT(self):
        S = StudentT([3, 1])
        self.assertFalse(S._check_input([3, -1]))

    def test_MultivariateNormal(self):
        M = MultivariateNormal([[1, 0], [[0.1, 0], [0, 0.1]]])
        self.assertFalse(M._check_input([[1, 0], [[1, 1], [0, 1]]]))

        self.assertFalse(M._check_input([[1, 0], [[-1, 0], [0, -1]]]))

    def test_MultiStudentT(self):
        M = MultiStudentT([[1, 0], [[1, 0], [0, 1]], 1])

        self.assertFalse(M._check_input([[1, 0], [[1, 1], [1, 0]], 1]))

        self.assertFalse(M._check_input([[1, 0], [[-1, 0], [0, -1]], 1]))

        self.assertFalse(M._check_input([[1, 0], [[1, 0], [0, 1]], -1]))


if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod

import numpy as np
from scipy.special import gamma
from scipy.stats import multivariate_normal

from abcpy.probabilisticmodels import Continuous


class PerturbationKernel(metaclass = ABCMeta):
    """This abstract base class represents all perturbation kernels"""

    @abstractmethod
    def __init__(self, models):
        """
        Parameters
        ----------
        models: list
            The list of abcpy.probabilisticmodel objects that should be perturbed by this kernel.
        """

        raise NotImplementedError


    @abstractmethod
    def calculate_cov(self, accepted_parameters_manager, kernel_index):
        """
        Calculates the covariance matrix for the kernel.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.acceptedparametersmanager object
            The accepted parameters manager that manages all bds objects.
        kernel_index: integer
            The index of the kernel in the list of kernels of the joint perturbation kernel.

        Returns
        -------
        numpy.ndarray:
            The covariance matrix for the kernel.
        """

        raise NotImplementedError


    @abstractmethod
    def update(self, accepted_parameters_manager, row_index, rng):
        """
        Perturbs the parameters for this kernel.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.acceptedparametersmanager object
            The accepted parameters manager that manages all bds objects.
        row_index: integer
            The index of the accepted parameters bds that should be perturbed.
        rng: random number generator
            The random number generator to be used.

        Returns
        -------
        numpy.ndarray:
            The perturbed parameters.
        """

        raise NotImplementedError


    def pdf(self, accepted_parameters_manager, kernel_index, row_index, x):
        """
        Calculates the pdf of the kernel at point x.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.acceptedparametersmanager object
            The accepted parameters manager that manages all bds objects.
        kernel_index: integer
            The index of the kernel in the list of kernels of the joint perturbation kernel.
        row_index: integer
            The index of the accepted parameters bds for which the pdf should be evaluated.
        x: list or float
            The point at which the pdf should be evaluated.

        Returns
        -------
        float:
            The pdf evaluated at point x.

        """

        if(isinstance(self, DiscreteKernel)):
            return self.pmf(accepted_parameters_manager, kernel_index, row_index, x)
        else:
            raise NotImplementedError


class ContinuousKernel(metaclass = ABCMeta):
    """This abstract base class represents all perturbation kernels acting on continuous parameters."""

    @abstractmethod
    def pdf(self, accepted_parameters_manager, kernel_index, index, x):
        raise NotImplementedError


class DiscreteKernel(metaclass = ABCMeta):
    """This abstract base class represents all perturbation kernels acting on discrete parameters."""

    @abstractmethod
    def pmf(self, accepted_parameters_manager, kernel_index, index, x):
        raise NotImplementedError


class JointPerturbationKernel(PerturbationKernel):
    def __init__(self, kernels):
        """
        This class joins different kernels to make up the overall perturbation kernel. Any user-implemented
        perturbation kernel should derive from this class. Any kernels defined on their own should be joined in the end
        using this class.

        Parameters
        ----------
        kernels: list
            List of abcpy.PerturbationKernels
        """

        self._check_kernels(kernels)
        self.kernels = kernels


    def calculate_cov(self, accepted_parameters_manager):
        """
        Calculates the covariance matrix corresponding to each kernel. Commonly used before calculating weights to avoid
        repeated calculation.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
             The AcceptedParametersManager to be uesd.

        Returns
        -------
        list
            Each entry corresponds to the covariance matrix of the corresponding kernel.
        """

        all_covs = []
        for kernel_index, kernel in enumerate(self.kernels):
            all_covs.append(kernel.calculate_cov(accepted_parameters_manager, kernel_index))
        return all_covs


    def _check_kernels(self, kernels):
        """
        Checks whether each model is only used in one perturbation kernel. Commonly called from the constructor.

        Parameters
        ----------
        kernels: list
            List of abcpy.PertubationKernels
        """

        models = []
        for kernel in kernels:
            for model in kernel.models:
                for already_contained_model in models:
                    if(already_contained_model==model):
                        raise ValueError("No two kernels can perturb the same probabilistic model.")
                models.append(model)


    def update(self, accepted_parameters_manager, row_index, rng=np.random.RandomState()):
        """
        Perturbs the parameter values contained in accepted_parameters_manager. Commonly used while perturbing.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
            Defines the AcceptedParametersManager to be used.
        row_index: integer
            The index of the row that should be considered from the accepted_parameters_bds matrix.
        rng: random number generator
            The random number generator to be used.

        Returns
        -------
        list
            The list contains tupels. Each tupel contains as the first entry a probabilistic model and as the second
            entry the perturbed parameter values corresponding to this model.
        """

        perturbed_values = []

        # Perturb values according to each kernel defined
        for kernel_index, kernel in enumerate(self.kernels):
            perturbed_values.append(kernel.update(accepted_parameters_manager, kernel_index, row_index, rng=rng))

        perturbed_values_including_models = []

        # Match the results from the perturbations and their models
        for kernel_index, kernel in enumerate(self.kernels):
            index=0
            for model in kernel.models:
                model_values = []
                #for j in range(model.get_output_dimension()):
                model_values.append(perturbed_values[kernel_index][index])
                index+=1
                perturbed_values_including_models.append((model, model_values))

        return perturbed_values_including_models


    def pdf(self, mapping, accepted_parameters_manager, mean, x):
        """
        Calculates the overall pdf of the kernel. Commonly used to calculate weights.

        Parameters
        ----------
        mapping: list
            Each entry is a tupel of which the first entry is a abcpy.ProbabilisticModel object, the second entry is the
            index in the accepted_parameters_bds list corresponding to an output of this model.
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
            The AcceptedParametersManager to be used.
        index: integer
            The row to be considered in the accepted_parameters_bds matrix.
        x: The point at which the pdf should be evaluated.

        Returns
        -------
        float
            The pdf evaluated at point x.
        """

        result = 1.
        for kernel_index, kernel in enumerate(self.kernels):
            # Define a list containing the parameter values relevant to the current kernel
            mean_kernel, theta = [], []
            for kernel_model in kernel.models:
                for model, model_output_index in mapping:
                    if(kernel_model==model):
                        theta.append(x[model_output_index])
                        mean_kernel.append(mean[model_output_index])
            result*=kernel.pdf(accepted_parameters_manager, kernel_index, mean_kernel, theta)

        return result


class MultivariateNormalKernel(PerturbationKernel, ContinuousKernel):
    """This class defines a kernel perturbing the parameters using a multivariate normal distribution."""

    def __init__(self, models):
        self.models = models


    def calculate_cov(self, accepted_parameters_manager, kernel_index):
        """
        Calculates the covariance matrix relevant to this kernel.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
            AcceptedParametersManager to be used.
        kernel_index: integer
            The index of the kernel in the list of kernels of the joint kernel.

        Returns
        -------
        list
            The covariance matrix corresponding to this kernel.
        """
        continuous_model = [[] for i in
                            range(len(accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index]))]
        for i in range(len(accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index])):
            if isinstance(accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index][i][0],
                          (np.float, np.float32, np.float64, np.int, np.int32, np.int64)):
                continuous_model[i] = accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index][i]
            else:
                continuous_model[i] = np.concatenate(
                    accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index][i])
        continuous_model = np.array(continuous_model).astype(float)

        if(accepted_parameters_manager.accepted_weights_bds is not None):
            weights = accepted_parameters_manager.accepted_weights_bds.value()
            cov = np.cov(continuous_model, aweights=weights.reshape(-1).astype(float), rowvar=False)
        else:
            cov = np.cov(continuous_model, rowvar=False)
        return cov


    def update(self, accepted_parameters_manager, kernel_index, row_index, rng=np.random.RandomState()):
        """
        Updates the parameter values contained in the accepted_paramters_manager using a multivariate normal distribution.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
            Defines the AcceptedParametersManager to be used.
        kernel_index: integer
            The index of the kernel in the list of kernels in the joint kernel.
        row_index: integer
            The index of the row that should be considered from the accepted_parameters_bds matrix.
        rng: random number generator
            The random number generator to be used.

        Returns
        -------
        np.ndarray
            The perturbed parameter values.
        """

        # Get all current parameter values relevant for this model and the structure
        continuous_model_values = accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index]

        if isinstance(continuous_model_values[row_index][0], (np.float, np.float32, np.float64, np.int, np.int32, np.int64)):
            # Perturb
            cov = np.array(accepted_parameters_manager.accepted_cov_mats_bds.value()[kernel_index]).astype(float)
            continuous_model_values = np.array(continuous_model_values).astype(float)

            # Perturbed values anc split according to the structure
            perturbed_continuous_values = rng.multivariate_normal(continuous_model_values[row_index], cov)
        else:
            #print('Hello')
            # Learn the structure
            struct = [[] for i in range(len(continuous_model_values[row_index]))]
            for i in range(len(continuous_model_values[row_index])):
                struct[i] = continuous_model_values[row_index][i].shape[0]
            struct = np.array(struct).cumsum()
            continuous_model_values = np.concatenate(continuous_model_values[row_index])

            # Perturb
            cov = np.array(accepted_parameters_manager.accepted_cov_mats_bds.value()[kernel_index]).astype(float)
            continuous_model_values = np.array(continuous_model_values).astype(float)

            # Perturbed values anc split according to the structure
            perturbed_continuous_values = np.split(rng.multivariate_normal(continuous_model_values, cov), struct)[:-1]

        return perturbed_continuous_values


    def pdf(self, accepted_parameters_manager, kernel_index, mean, x):
        """Calculates the pdf of the kernel.
        Commonly used to calculate weights.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
            The AcceptedParametersManager to be used.
        kernel_index: integer
            The index of the kernel in the list of kernels in the joint kernel.
        index: integer
            The row to be considered in the accepted_parameters_bds matrix.
        x: The point at which the pdf should be evaluated.

        Returns
        -------
        float
            The pdf evaluated at point x.
        """

        if isinstance(mean[0], (np.float, np.float32, np.float64, np.int, np.int32, np.int64)):
            mean = np.array(mean).astype(float)
            cov = np.array(accepted_parameters_manager.accepted_cov_mats_bds.value()[kernel_index]).astype(float)
            return multivariate_normal(mean, cov, allow_singular=True).pdf(np.array(x).astype(float))
        else:
            mean = np.array(np.concatenate(mean)).astype(float)
            cov = np.array(accepted_parameters_manager.accepted_cov_mats_bds.value()[kernel_index]).astype(float)
            return multivariate_normal(mean, cov, allow_singular=True).pdf(np.concatenate(x))


class MultivariateStudentTKernel(PerturbationKernel, ContinuousKernel):
    def __init__(self, models, df):
        """This class defines a kernel perturbing the parameters using a multivariate normal distribution.

        Parameters
        ----------
        models: list of abcpy.probabilisticmodel objects
            The models that should be perturbed using this kernel
        df: integer
            The degrees of freedom to be used.
        """

        self.models = models
        self.df = df


    def calculate_cov(self, accepted_parameters_manager, kernel_index):
        """
        Calculates the covariance matrix relevant to this kernel.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
            AcceptedParametersManager to be used.
        kernel_index: integer
            The index of the kernel in the list of kernels of the joint kernel.

        Returns
        -------
        list
            The covariance matrix corresponding to this kernel.
        """
        continuous_model = [[] for i in
                            range(len(accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index]))]
        for i in range(len(accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index])):
            if isinstance(accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index][i][0],
                          (np.float, np.float32, np.float64, np.int, np.int32, np.int64)):
                continuous_model[i] = accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index][i]
            else:
                continuous_model[i] = np.concatenate(
                    accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index][i])
        continuous_model = np.array(continuous_model).astype(float)

        if(accepted_parameters_manager.accepted_weights_bds is not None):
            weights = np.array(accepted_parameters_manager.accepted_weights_bds.value())
            cov = np.cov(continuous_model, aweights=weights.reshape(-1).astype(float),rowvar=False)
        else:
            cov = np.cov(continuous_model, rowvar=False)
        return cov


    def update(self, accepted_parameters_manager, kernel_index, row_index, rng=np.random.RandomState()):
        """
        Updates the parameter values contained in the accepted_paramters_manager using a multivariate normal distribution.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
            Defines the AcceptedParametersManager to be used.
        kernel_index: integer
            The index of the kernel in the list of kernels in the joint kernel.
        row_index: integer
            The index of the row that should be considered from the accepted_parameters_bds matrix.
        rng: random number generator
            The random number generator to be used.

        Returns
        -------
        np.ndarray
            The perturbed parameter values.

        """

        # Get all parameters relevant to this kernel
        continuous_model_values = accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index][row_index]

        if isinstance(continuous_model_values[0],
                      (np.float, np.float32, np.float64, np.int, np.int32, np.int64)):
            # Perturb
            continuous_model_values = np.array(continuous_model_values)
            cov = np.array(accepted_parameters_manager.accepted_cov_mats_bds.value()[kernel_index]).astype(float)
            p = len(continuous_model_values)

            if (self.df == np.inf):
                chisq = 1.0
            else:
                chisq = rng.chisquare(self.df, 1) / self.df
                chisq = chisq.reshape(-1, 1).repeat(p, axis=1)

            mvn = rng.multivariate_normal(np.zeros(p), cov.astype(float), 1)
            perturbed_continuous_values = continuous_model_values + np.divide(mvn, np.sqrt(chisq))[0]
        else:
            # Learn the structure
            struct = [[] for i in range(len(continuous_model_values))]
            for i in range(len(continuous_model_values)):
                struct[i] = continuous_model_values[i].shape[0]
            struct = np.array(struct).cumsum()
            continuous_model_values = np.concatenate(continuous_model_values)

            # Perturb
            cov = np.array(accepted_parameters_manager.accepted_cov_mats_bds.value()[kernel_index]).astype(float)
            p = len(continuous_model_values)

            if (self.df == np.inf):
                chisq = 1.0
            else:
                chisq = rng.chisquare(self.df, 1) / self.df
                chisq = chisq.reshape(-1, 1).repeat(p, axis=1)

            mvn = rng.multivariate_normal(np.zeros(p), cov.astype(float), 1)
            perturbed_continuous_values = continuous_model_values + np.divide(mvn, np.sqrt(chisq))[0]

            # Perturbed values anc split according to the structure
            perturbed_continuous_values = np.split(perturbed_continuous_values, struct)[:-1]

        return perturbed_continuous_values

    def pdf(self, accepted_parameters_manager, kernel_index, mean, x):
        """Calculates the pdf of the kernel.
        Commonly used to calculate weights.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
            The AcceptedParametersManager to be used.
        kernel_index: integer
            The index of the kernel in the list of kernels in the joint kernel.
        index: integer
            The row to be considered in the accepted_parameters_bds matrix.
        x: The point at which the pdf should be evaluated.

        Returns
        -------
        float
            The pdf evaluated at point x.
        """

        if isinstance(mean[0],
                      (np.float, np.float32, np.float64, np.int, np.int32, np.int64)):
            mean = np.array(mean).astype(float)
            cov = np.array(accepted_parameters_manager.accepted_cov_mats_bds.value()[kernel_index]).astype(float)

            v = self.df
            p = len(mean)

            numerator = gamma((v + p) / 2)
            denominator = gamma(v / 2) * pow(v * np.pi, p / 2.) * np.sqrt(abs(np.linalg.det(cov)))
            normalizing_const = numerator / denominator
            tmp = 1 + 1 / v * np.dot(np.dot(np.transpose(x - mean), np.linalg.inv(cov)), (x - mean))
            density = normalizing_const * pow(tmp, -((v + p) / 2.))

            return density
        else:
            mean = np.array(np.concatenate(mean)).astype(float)
            cov = np.array(accepted_parameters_manager.accepted_cov_mats_bds.value()[kernel_index]).astype(float)

            v = self.df
            p = len(mean)

            numerator = gamma((v + p) / 2)
            denominator = gamma(v / 2) * pow(v * np.pi, p / 2.) * np.sqrt(abs(np.linalg.det(cov)))
            normalizing_const = numerator / denominator
            tmp = 1 + 1 / v * np.dot(np.dot(np.transpose(np.concatenate(x) - mean), np.linalg.inv(cov)), (np.concatenate(x) - mean))
            density = normalizing_const * pow(tmp, -((v + p) / 2.))

            return density

class RandomWalkKernel(PerturbationKernel, DiscreteKernel):
    def __init__(self, models):
        """
        This class defines a kernel perturbing discrete parameters using a naive random walk.

        Parameters
        ----------
        models: list
            List of abcpy.ProbabilisticModel objects
        """

        self.models = models

    def update(self, accepted_parameters_manager, kernel_index, row_index, rng=np.random.RandomState()):
        """
        Updates the parameter values contained in the accepted_paramters_manager using a random walk.

        Parameters
        ----------
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
            Defines the AcceptedParametersManager to be used.
        row_index: integer
            The index of the row that should be considered from the accepted_parameters_bds matrix.
        rng: random number generator
            The random number generator to be used.

        Returns
        -------
        np.ndarray
            The perturbed parameter values.

        """

        # Get parameter values relevant to this kernel
        discrete_model_values = accepted_parameters_manager.kernel_parameters_bds.value()[kernel_index]

        perturbed_discrete_values = []
        discrete_model_values = np.array(discrete_model_values)[row_index]

        # Implement a random walk for the discrete parameter values
        for discrete_value in discrete_model_values:
            perturbed_discrete_values.append(np.array([rng.randint(discrete_value - 1, discrete_value + 2)]))

        return perturbed_discrete_values


    def calculate_cov(self, accepted_parameters_manager, kernel_index):
        """
        Calculates the covariance matrix of this kernel. Since there is no covariance matrix associated with this
        random walk, it returns an empty list.
        """

        return np.array([0]).reshape(-1,)


    def pmf(self, accepted_parameters_manager, kernel_index, mean, x):
        """
        Calculates the pmf of the kernel. Commonly used to calculate weights.

        Parameters
        ----------
        cov: list
            The covariance matrix used for this kernel. This is a dummy input.
        accepted_parameters_manager: abcpy.AcceptedParametersManager object
            The AcceptedParametersManager to be used.
        kernel_index: integer
            The index of the kernel in the list of kernels of the joint kernel.
        index: integer
            The row to be considered in the accepted_parameters_bds matrix.
        x: The point at which the pdf should be evaluated.

        Returns
        -------
        float
            The pmf evaluated at point x.
        """

        return 1./3


class DefaultKernel(JointPerturbationKernel):
    def __init__(self, models):
        """
        This class implements a kernel that perturbs all continuous parameters using a multivariate normal, and all
        discrete parameters using a random walk. To be used as an example for user defined kernels.

        Parameters
        ----------
        models: list
            List of abcpy.ProbabilisticModel objects, the models for which the kernel should be defined.
        """

        continuous_models = []
        discrete_models = []
        for model in models:
            if(isinstance(model, Continuous)):
                continuous_models.append(model)
            else:
                discrete_models.append(model)
        continuous_kernel = MultivariateNormalKernel(continuous_models)
        discrete_kernel = RandomWalkKernel(discrete_models)
        if(not(continuous_models)):
            super(DefaultKernel, self).__init__([discrete_kernel])
        elif(not(discrete_models)):
            super(DefaultKernel, self).__init__([continuous_kernel])
        else:
            super(DefaultKernel, self).__init__([continuous_kernel, discrete_kernel])

